from collections import deque
from typing import List, Iterable, Tuple, Optional

import numpy as np

from golem.core.optimisers.adaptive.common_types import ObsType, ActType, TrajectoryStep, GraphTrajectory
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory


class ExperienceBuffer:
    """Buffer for learning experience of ``OperatorAgent``.
    Keeps (State, Action, Reward) lists until retrieval."""

    def __init__(self, window_size: Optional[int] = None, inds=None, actions=None, rewards=None):
        self.window_size = window_size
        self._prev_pop = set()
        self._next_pop = set()

        if inds and not (len(inds) == len(actions) == len(rewards)):
            raise ValueError('lengths of buffers do not match')
        self._individuals = deque(inds) if inds else deque(maxlen=self.window_size)
        self._actions = deque(actions) if actions else deque(maxlen=self.window_size)
        self._rewards = deque(rewards) if rewards else deque(maxlen=self.window_size)

    @staticmethod
    def from_history(history: OptHistory) -> 'ExperienceBuffer':
        exp = ExperienceBuffer()
        exp.collect_history(history)
        return exp

    def _reset(self):
        self._prev_pop = set()
        self._next_pop = set()

        # if window size was not specified than there is no need to store these values for reuse.
        # Otherwise, if the window_size is specified, then storages will be updated automatically in queue
        if self.window_size is None:
            self._individuals = deque(maxlen=self.window_size)
            self._actions = deque(maxlen=self.window_size)
            self._rewards = deque(maxlen=self.window_size)

    @staticmethod
    def unroll_action_step(result: Individual) -> TrajectoryStep:
        """Unrolls individual's history to get its source individual, action and resulting reward."""
        if not result.parent_operator or result.parent_operator.type_ != 'mutation':
            return None, None, np.nan
        source_ind = result.parent_operator.parent_individuals[0]
        action = result.parent_operator.operators[0]
        # we're minimising the fitness, that's why less is better
        reward = (source_ind.fitness.value - result.fitness.value) / abs(source_ind.fitness.value)\
            if source_ind.fitness and source_ind.fitness.value != 0. else 0.
        return source_ind, action, reward

    @staticmethod
    def unroll_trajectories(history: OptHistory) -> List[GraphTrajectory]:
        """Iterates through history and find continuous sequences of applied operator actions."""
        trajectories = []
        seen_uids = set()
        for terminal_individual in history.final_choices:
            trajectory = []
            next_ind = terminal_individual
            while True:
                seen_uids.add(next_ind.uid)
                source_ind, action, reward = ExperienceBuffer.unroll_action_step(next_ind)
                if source_ind is None or source_ind.uid in seen_uids:
                    break
                # prepend step to keep historical direction
                trajectory.insert(0, (source_ind, action, reward))
                next_ind = source_ind
            trajectories.append(trajectory)
        return trajectories

    def collect_history(self, history: OptHistory):
        seen = set()
        # We don't need the initial assumptions, as they have no parent operators, hence [1:]
        for generation in history.generations[1:]:
            for ind in generation:
                if ind.uid not in seen:
                    seen.add(ind.uid)
                    self.collect_result(ind)

    def collect_results(self, results: Iterable[Individual]):
        for ind in results:
            self.collect_result(ind)

    def collect_result(self, result: Individual):
        if result.uid in self._prev_pop:
            # avoid collecting results from individuals that didn't change
            return
        self._next_pop.add(result.uid)

        source_ind, action, reward = self.unroll_action_step(result)
        if action is None:
            return
        self.collect_experience(source_ind, action, reward)

    def collect_experience(self, obs: Individual, action: ActType, reward: float):
        self._individuals.append(obs)
        self._actions.append(action)
        self._rewards.append(reward)

    def retrieve_experience(self, as_graphs: bool = True) -> Tuple[List[ObsType], List[ActType], List[float]]:
        """Get all collected experience and clear the experience buffer.
        Args:
            as_graphs: if True (by default) returns observations as graphs, otherwise as individuals.
        Return:
             Unzipped trajectories (tuple of lists of observations, actions, rewards).
        """
        individuals, actions, rewards = self._individuals, self._actions, self._rewards
        observations = [ind.graph for ind in individuals] if as_graphs else individuals
        next_pop = self._next_pop
        self._reset()
        self._prev_pop = next_pop
        return list(observations), list(actions), list(rewards)

    def retrieve_trajectories(self) -> GraphTrajectory:
        """Same as `retrieve_experience` but in the form of zipped trajectories that consist from steps."""
        trajectories = list(zip(*self.retrieve_experience(as_graphs=False)))
        return trajectories

    def split(self, ratio: float = 0.8, shuffle: bool = False
              ) -> Tuple['ExperienceBuffer', 'ExperienceBuffer']:
        """Splits buffer in 2 parts, useful for train/validation split."""
        mask_train = np.full_like(self._individuals, False, dtype=bool)
        num_train = int(len(self._individuals) * ratio)
        mask_train[-num_train:] = True
        if shuffle:
            np.random.default_rng().shuffle(mask_train)
        buffer_train = ExperienceBuffer(inds=np.array(self._individuals)[mask_train].tolist(),
                                        actions=np.array(self._actions)[mask_train].tolist(),
                                        rewards=np.array(self._rewards)[mask_train].tolist())
        buffer_val = ExperienceBuffer(inds=np.array(self._individuals)[~mask_train].tolist(),
                                      actions=np.array(self._actions)[~mask_train].tolist(),
                                      rewards=np.array(self._rewards)[~mask_train].tolist())
        return buffer_train, buffer_val

    def __len__(self):
        return len(self._individuals)

    def __str__(self):
        return f'{self.__class__.__name__}({len(self)})'



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Optional, Sequence

from golem.core.log import default_log
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.opt_history_objects.individual import Individual

if TYPE_CHECKING:
    from golem.core.optimisers.optimizer import AlgorithmParameters

PopulationT = Sequence[Individual]
EvaluationOperator = Callable[[PopulationT], PopulationT]


class Operator(ABC):
    """ Base abstract functional interface for genetic operators.
    Specific signatures are:
    - Selection: Population -> Population
    - Inheritance: [Population, Population] -> Population
    - Regularization: [Population, EvaluationOperator] -> Population
    - Mutation: Union[Individual, Population] -> Union[Individual, Population]
    - Crossover: Population -> Population
    - Elitism: [Population, Population] -> Population
    """

    def __init__(self,
                 parameters: Optional['AlgorithmParameters'] = None,
                 requirements: Optional[GraphRequirements] = None):
        self.requirements = requirements
        self.parameters = parameters
        self.log = default_log(self)

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    def update_requirements(self,
                            parameters: Optional['AlgorithmParameters'] = None,
                            requirements: Optional[GraphRequirements] = None):
        if requirements:
            self.requirements = requirements
        if parameters:
            self.parameters = parameters



from abc import abstractmethod
from random import choice
from typing import Any, Optional, Sequence, Dict

from golem.core.constants import MIN_POP_SIZE
from golem.core.dag.graph import Graph
from golem.core.optimisers.archive import GenerationKeeper
from golem.core.optimisers.genetic.evaluation import MultiprocessingDispatcher, SequentialDispatcher
from golem.core.optimisers.genetic.operators.operator import PopulationT, EvaluationOperator
from golem.core.optimisers.objective import GraphFunction, ObjectiveFunction
from golem.core.optimisers.objective.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer, AlgorithmParameters
from golem.core.optimisers.timer import OptimisationTimer
from golem.utilities.grouped_condition import GroupedCondition


class PopulationalOptimizer(GraphOptimizer):
    """
    Base class of populational optimizer.
    PopulationalOptimizer implements all basic methods for optimization not related to evolution process
    to experiment with other kinds of evolution optimization methods
    It allows to find the optimal solution using specified metric (one or several).
    To implement the specific evolution strategy, implement `_evolution_process`.

    Args:
         objective: objective for optimization
         initial_graphs: graphs which were initialized outside the optimizer
         requirements: implementation-independent requirements for graph optimizer
         graph_generation_params: parameters for new graph generation
         graph_optimizer_params: parameters for specific implementation of graph optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[Graph],
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams,
                 graph_optimizer_params: Optional['AlgorithmParameters'] = None,
                 ):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.population = None
        self.generations = GenerationKeeper(self.objective, keep_n_best=requirements.keep_n_best)
        self.timer = OptimisationTimer(timeout=self.requirements.timeout)

        dispatcher_type = MultiprocessingDispatcher if self.requirements.parallelization_mode == 'populational' else \
            SequentialDispatcher

        self.eval_dispatcher = dispatcher_type(adapter=graph_generation_params.adapter,
                                               n_jobs=requirements.n_jobs,
                                               graph_cleanup_fn=_try_unfit_graph,
                                               delegate_evaluator=graph_generation_params.remote_evaluator)

        # early_stopping_iterations and early_stopping_timeout may be None, so use some obvious max number
        max_stagnation_length = requirements.early_stopping_iterations or requirements.num_of_generations
        max_stagnation_time = requirements.early_stopping_timeout or self.timer.timeout
        self.stop_optimization = \
            GroupedCondition(results_as_message=True).add_condition(
                lambda: self.timer.is_time_limit_reached(self.current_generation_num - 1),
                'Optimisation stopped: Time limit is reached'
            ).add_condition(
                lambda: (requirements.num_of_generations is not None and
                         self.current_generation_num >= requirements.num_of_generations + 1),
                'Optimisation stopped: Max number of generations reached'
            ).add_condition(
                lambda: (max_stagnation_length is not None and
                         self.generations.stagnation_iter_count >= max_stagnation_length),
                'Optimisation finished: Early stopping iterations criteria was satisfied'
            ).add_condition(
                lambda: self.generations.stagnation_time_duration >= max_stagnation_time,
                'Optimisation finished: Early stopping timeout criteria was satisfied'
            )
        # in how many generations structural diversity check should be performed
        self.gen_structural_diversity_check = self.graph_optimizer_params.structural_diversity_frequency_check

    @property
    def current_generation_num(self) -> int:
        return self.generations.generation_num

    def set_evaluation_callback(self, callback: Optional[GraphFunction]):
        # Redirect callback to evaluation dispatcher
        self.eval_dispatcher.set_graph_evaluation_callback(callback)

    def optimise(self, objective: ObjectiveFunction) -> Sequence[Graph]:

        # eval_dispatcher defines how to evaluate objective on the whole population
        evaluator = self.eval_dispatcher.dispatch(objective, self.timer)

        with self.timer, self._progressbar as pbar:

            self._initial_population(evaluator)

            while not self.stop_optimization():
                try:
                    new_population = self._evolve_population(evaluator)
                    if self.gen_structural_diversity_check != -1 \
                            and self.generations.generation_num % self.gen_structural_diversity_check == 0 \
                            and self.generations.generation_num != 0:
                        new_population = self.get_structure_unique_population(new_population, evaluator)
                    pbar.update()
                except EvaluationAttemptsError as ex:
                    self.log.warning(f'Composition process was stopped due to: {ex}')
                    break
                # Adding of new population to history
                self._update_population(new_population)
        pbar.close()
        self._update_population(self.best_individuals, 'final_choices')
        return [ind.graph for ind in self.best_individuals]

    @property
    def best_individuals(self):
        return self.generations.best_individuals

    @abstractmethod
    def _initial_population(self, evaluator: EvaluationOperator):
        """ Initializes the initial population """
        raise NotImplementedError()

    @abstractmethod
    def _evolve_population(self, evaluator: EvaluationOperator) -> PopulationT:
        """ Method realizing full evolution cycle """
        raise NotImplementedError()

    def _extend_population(self, pop: PopulationT, target_pop_size: int) -> PopulationT:
        """ Extends population to specified `target_pop_size`. """
        n = target_pop_size - len(pop)
        extended_population = list(pop)
        extended_population.extend([Individual(graph=choice(pop).graph) for _ in range(n)])
        return extended_population

    def _update_population(self, next_population: PopulationT, label: Optional[str] = None,
                           metadata: Optional[Dict[str, Any]] = None):
        self.generations.append(next_population)
        if self.requirements.keep_history:
            self._log_to_history(next_population, label, metadata)
        self._iteration_callback(next_population, self)
        self.population = next_population

        self.log.info(f'Generation num: {self.current_generation_num} size: {len(next_population)}')
        self.log.info(f'Best individuals: {str(self.generations)}')
        if self.generations.stagnation_iter_count > 0:
            self.log.info(f'no improvements for {self.generations.stagnation_iter_count} iterations')
            self.log.info(f'spent time: {round(self.timer.minutes_from_start, 1)} min')

    def _log_to_history(self, population: PopulationT, label: Optional[str] = None,
                        metadata: Optional[Dict[str, Any]] = None):
        self.history.add_to_history(population, label, metadata)
        self.history.add_to_archive_history(self.generations.best_individuals)
        if self.requirements.history_dir:
            self.history.save_current_results(self.requirements.history_dir)

    def get_structure_unique_population(self, population: PopulationT, evaluator: EvaluationOperator) -> PopulationT:
        """ Increases structurally uniqueness of population to prevent stagnation in optimization process.
        Returned population may be not entirely unique, if the size of unique population is lower than MIN_POP_SIZE. """
        unique_population_with_ids = {ind.graph.descriptive_id: ind for ind in population}
        unique_population = list(unique_population_with_ids.values())

        # if size of unique population is too small, then extend it to MIN_POP_SIZE by repeating individuals
        if len(unique_population) < MIN_POP_SIZE:
            unique_population = self._extend_population(pop=unique_population, target_pop_size=MIN_POP_SIZE)
        return evaluator(unique_population)


# TODO: remove this hack (e.g. provide smth like FitGraph with fit/unfit interface)
def _try_unfit_graph(graph: Any):
    if hasattr(graph, 'unfit'):
        graph.unfit()


class EvaluationAttemptsError(Exception):
    """ Number of evaluation attempts exceeded """

    def __init__(self, *args):
        self.message = args[0] or None

    def __str__(self):
        return self.message or 'Too many fitness evaluation errors.'



import random
from abc import ABC
from copy import deepcopy
from os import makedirs
from os.path import exists, join
from typing import List, Optional, Type, Union, Dict, Callable

from golem.core.log import default_log
from golem.core.dag.graph import Graph
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.timer import OptimisationTimer
from golem.core.paths import default_data_dir
from golem.structural_analysis.base_sa_approaches import BaseAnalyzeApproach
from golem.structural_analysis.graph_sa.entities.edge import Edge
from golem.structural_analysis.graph_sa.results.deletion_sa_approach_result import \
    DeletionSAApproachResult
from golem.structural_analysis.graph_sa.results.object_sa_result import ObjectSAResult
from golem.structural_analysis.graph_sa.results.replace_sa_approach_result import \
    ReplaceSAApproachResult
from golem.structural_analysis.graph_sa.results.utils import EntityTypesEnum
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements, \
    ReplacementAnalysisMetaParams


class EdgeAnalysis:
    """
    Class for successively applying approaches for structural analysis

    :param approaches: methods applied to edges to modify the graph or analyze certain operations.\
    Default: [EdgeDeletionAnalyze, EdgeReplaceOperationAnalyze]
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, approaches: Optional[List[Type['EdgeAnalyzeApproach']]] = None,
                 approaches_requirements: Optional[StructuralAnalysisRequirements] = None,
                 path_to_save: Optional[str] = None):

        self.approaches = [EdgeDeletionAnalyze, EdgeReplaceOperationAnalyze] \
            if approaches is None else approaches

        self.path_to_save = \
            join(default_data_dir(), 'structural', 'edges_structural') if path_to_save is None else path_to_save
        self.log = default_log(self)

        self.approaches_requirements = \
            StructuralAnalysisRequirements() if approaches_requirements is None else approaches_requirements

    def analyze(self, graph: Graph, edge: Edge,
                objective: Callable,
                timer: Optional[OptimisationTimer] = None) -> ObjectSAResult:
        """
        Method runs Edge analysis within defined approaches

        :param graph: graph containing the analyzed Edge
        :param edge: Edge object to analyze in Graph
        :param objective: objective function for computing metric values
        :param timer: timer to check if the time allotted for structural analysis has expired
        :return: dict with Edge analysis result per approach
        """

        results = ObjectSAResult(entity_idx=
                                 f'{graph.nodes.index(edge.parent_node)}_{graph.nodes.index(edge.child_node)}',
                                 entity_type=EntityTypesEnum.edge)

        for approach in self.approaches:
            if timer is not None and timer.is_time_limit_reached():
                break

            results.add_result(approach(graph=graph,
                                        objective=objective,
                                        requirements=self.approaches_requirements,
                                        path_to_save=self.path_to_save).analyze(edge=edge))

        return results


class EdgeAnalyzeApproach(BaseAnalyzeApproach, ABC):
    """
    Base class for edge analysis approach.

    :param graph: Graph containing the analyzing Edge
    :param objective: objective function for computing metric values
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, graph: Graph, objective: Objective, requirements: StructuralAnalysisRequirements = None,
                 path_to_save=None):
        super().__init__(graph, objective, requirements)
        self._origin_metrics = None

        self._path_to_save = \
            join(default_data_dir(), 'structural', 'edges_structural') if path_to_save is None else path_to_save
        self.log = default_log(prefix='edge_analysis')

        if not exists(self._path_to_save):
            makedirs(self._path_to_save)


class EdgeDeletionAnalyze(EdgeAnalyzeApproach):
    def __init__(self, graph: Graph, objective: Objective,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, requirements)

        self._path_to_save = \
            join(default_data_dir(), 'structural', 'edges_structural') if path_to_save is None else path_to_save
        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    def analyze(self, edge: Edge, **kwargs) -> DeletionSAApproachResult:
        """
        Receives a graph without the specified edge and tries to calculate the loss for it

        :param edge: Edge object to analyze
        :return: the ratio of modified graph score to origin score
        """
        results = DeletionSAApproachResult()
        shortened_graph = self.sample(edge)
        if shortened_graph:
            losses = self._compare_with_origin_by_metrics(shortened_graph)
            self.log.message(f'loss: {losses}')
            del shortened_graph
        else:
            self.log.warning('if remove this edge then get an invalid graph')
            losses = [-1.0] * len(self._objective.metrics)

        results.add_results(metrics_values=losses)
        return results

    def sample(self, edge: Edge) -> Optional[Graph]:
        """
        Checks if it is possible to delete an edge from the graph so that it remains valid,
        and if so, deletes

        :param edge: Edge object to delete from Graph object
        :return: Graph object without edge
        """

        graph_sample = deepcopy(self._graph)

        parent_node_index_to_delete = self._graph.nodes.index(edge.parent_node)
        parent_node_to_delete = graph_sample.nodes[parent_node_index_to_delete]

        child_node_index_to_delete = self._graph.nodes.index(edge.child_node)
        child_node_to_delete = graph_sample.nodes[child_node_index_to_delete]

        graph_sample.disconnect_nodes(parent_node_to_delete, child_node_to_delete)

        verifier = self._requirements.graph_verifier
        if not verifier.verify(graph_sample):
            self.log.warning('Can not delete edge since modified graph can not pass verification')
            return None

        return graph_sample


class EdgeReplaceOperationAnalyze(EdgeAnalyzeApproach):
    """
       Replace edge with operations available for the current task
       and evaluate the score difference
    """

    def __init__(self, graph: Graph, objective: Objective,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, requirements)

        self._path_to_save = \
            join(default_data_dir(), 'structural', 'edges_structural') if path_to_save is None else path_to_save
        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    def analyze(self, edge: Edge, **kwargs) -> ReplaceSAApproachResult:
        """
        Counts the loss on each changed graph received and returns the biggest loss and
        the graph on which it was received

        :param edge: Edge object to analyze
        :return: dictionary of the best (the biggest) loss and corresponding to it edge to replace to
        """
        result = ReplaceSAApproachResult()
        requirements: ReplacementAnalysisMetaParams = self._requirements.replacement_meta
        samples_res = self.sample(edge=edge,
                                  edges_idxs_to_replace_to=requirements.edges_to_replace_to,
                                  number_of_random_operations=requirements.number_of_random_operations_edges)

        samples = samples_res['samples']
        edges_nodes_idx_to_replace_to = samples_res['edges_nodes_idx_to_replace_to']

        loss_values = []
        for i, sample_graph in enumerate(samples):
            loss_per_sample = self._compare_with_origin_by_metrics(sample_graph)
            self.log.message(f'loss: {loss_per_sample}')
            loss_values.append(loss_per_sample)

            child_node_idx = ''
            parent_node_idx = ''
            part1, part2 = edges_nodes_idx_to_replace_to[i].__str__().split(',')
            for char in part1:
                if char.isdigit():
                    child_node_idx += char
                else:
                    continue
            for char in part2:
                if char.isdigit():
                    parent_node_idx += char
                else:
                    continue
            result.add_results(entity_to_replace_to=f'{parent_node_idx}_{child_node_idx}',
                               metrics_values=loss_per_sample)
        return result

    def sample(self, edge: Edge,
               edges_idxs_to_replace_to: Optional[List[Edge]],
               number_of_random_operations: Optional[int] = 1) \
            -> Dict[str, Union[List[Graph], List[Dict[str, int]]]]:
        """
        Tries to replace the given edge with a pool of edges available for replacement (see _edge_generation docstring)
        and validates the resulting graphs

        :param edge: Edge object to replace
        :param edges_idxs_to_replace_to: edges provided for old_edge replacement
        :param number_of_random_operations: number of replacement operations, \
        if edges_to_replace_to not provided

        :return: dictionary of sequence of Graph objects with new edges instead of old one
        and indexes of edges to which to change the given edge to get these graphs
        """

        if not edges_idxs_to_replace_to or number_of_random_operations:
            edges_idxs_to_replace_to = self._edge_generation(edge=edge,
                                                             number_of_operations=number_of_random_operations)
        samples = list()
        edges_nodes_idx_to_replace_to = list()
        for replacing_nodes_idx in edges_idxs_to_replace_to:
            sample_graph = deepcopy(self._graph)

            # disconnect nodes
            previous_parent_node_index = self._graph.nodes.index(edge.parent_node)
            previous_child_node_index = self._graph.nodes.index(edge.child_node)

            previous_parent_node = sample_graph.nodes[previous_parent_node_index]
            previous_child_node = sample_graph.nodes[previous_child_node_index]

            # connect nodes
            next_parent_node = sample_graph.nodes[replacing_nodes_idx['parent_node_idx']]
            next_child_node = sample_graph.nodes[replacing_nodes_idx['child_node_idx']]

            if next_parent_node in sample_graph.nodes and \
                    next_child_node in sample_graph.nodes:
                sample_graph.connect_nodes(next_parent_node, next_child_node)

            sample_graph.disconnect_nodes(node_parent=previous_parent_node,
                                          node_child=previous_child_node,
                                          clean_up_leftovers=False)

            verifier = self._requirements.graph_verifier
            if not verifier.verify(sample_graph):
                self.log.warning('Can not connect these nodes')
            else:
                self.log.message(f'replace edge parent: {next_parent_node}')
                self.log.message(f'replace edge child: {next_child_node}')
                samples.append(sample_graph)
                edges_nodes_idx_to_replace_to.append({'parent_node_id':
                                                     replacing_nodes_idx['parent_node_idx'],
                                                      'child_node_id':
                                                     replacing_nodes_idx['child_node_idx']})

        if not edges_nodes_idx_to_replace_to:
            res = {'samples': [self._graph], 'edges_nodes_idx_to_replace_to':
                [{'parent_node_id': self._graph.nodes.index(edge.parent_node),
                  'child_node_id': self._graph.nodes.index(edge.child_node)}]}
            return res

        return {'samples': samples, 'edges_nodes_idx_to_replace_to': edges_nodes_idx_to_replace_to}

    def _edge_generation(self, edge: Edge, number_of_operations: int = 1) -> List[Dict[str, int]]:
        """
        The method returns possible edges that can replace the given edge.
        These edges must not start at the root node, already exist in the graph and must not form cycles

        :param edge: edge to be replaced
        :param number_of_operations: limits the number of possible edges to replace to

        :return: edges with which it's possible to replace the passed edge
        """
        cur_graph = deepcopy(self._graph)

        parent_node_index = self._graph.nodes.index(edge.parent_node)
        child_node_index = self._graph.nodes.index(edge.child_node)

        parent_node = cur_graph.nodes[parent_node_index]
        child_node = cur_graph.nodes[child_node_index]

        if child_node is cur_graph.root_node and len(child_node.nodes_from) == 1:
            return []

        cur_graph.disconnect_nodes(node_parent=parent_node, node_child=child_node,
                                   clean_up_leftovers=False)

        edges_in_graph = cur_graph.get_edges()

        available_edges_idx = list()

        for parent_node in cur_graph.nodes[1:]:
            for child_node in cur_graph.nodes:
                if parent_node == child_node:
                    continue
                if [parent_node, child_node] in edges_in_graph or [child_node, parent_node] in edges_in_graph:
                    continue
                if cur_graph.nodes.index(parent_node) == child_node_index and \
                        cur_graph.nodes.index(child_node) == parent_node_index:
                    continue
                available_edges_idx.append({'parent_node_idx': cur_graph.nodes.index(parent_node),
                                            'child_node_idx': cur_graph.nodes.index(child_node)})

        edges_for_replacement = random.sample(available_edges_idx, min(number_of_operations, len(available_edges_idx)))
        return edges_for_replacement



import os
import random
import datetime

import numpy as np

from examples.graph_model_optimization import run_custom_example
from golem.utilities.profiler.memory_profiler import MemoryProfiler
from golem.utilities.profiler.time_profiler import TimeProfiler

random.seed(1)
np.random.seed(1)

if __name__ == '__main__':
    # JUST UNCOMMENT WHAT TYPE OF PROFILER DO YOU NEED
    # EXAMPLE of MemoryProfiler.

    arguments = dict(timeout=datetime.timedelta(minutes=0.5), visualisation=False)
    path = os.path.join(os.path.expanduser("~"), 'memory_profiler')
    MemoryProfiler(run_custom_example, kwargs=arguments,
                   path=path, roots=[run_custom_example], max_depth=8, visualization=True)

    # EXAMPLE of TimeProfiler.

    profiler = TimeProfiler()
    run_custom_example(timeout=datetime.timedelta(minutes=0.5), visualisation=False)
    path = os.path.join(os.path.expanduser("~"), 'time_profiler')
    profiler.profile(path=path, node_percent=0.5, edge_percent=0.1, open_web=True)



from typing import Dict, Callable, Collection, Sequence, Optional

import networkx as nx
import numpy as np

from examples.synthetic_graph_evolution.utils import draw_graphs_subplots

NumNodes = int
DiGraphGenerator = Callable[[NumNodes], nx.DiGraph]


graph_generators: Dict[str, DiGraphGenerator] = {
    'line': lambda n: nx.path_graph(n, create_using=nx.DiGraph),
    'tree': lambda n: nx.random_tree(n, create_using=nx.DiGraph),
    'gnp': lambda n: nx.gnp_random_graph(n, p=0.1),
    'dag': lambda n: generate_dag(n),
    'star': nx.star_graph,
    '2ring': nx.circular_ladder_graph,
    'grid2d': lambda n: nx.grid_2d_graph(int(np.sqrt(n)), int(np.sqrt(n))),
    'hypercube': lambda n: nx.hypercube_graph(int(np.log2(n).round())),
}

graph_kinds: Sequence[str] = tuple(graph_generators.keys())


def generate_dag(n):
    """ Works good for small graphs (up to n=100000) """
    g = nx.gnp_random_graph(n, p=0.5, directed=True)
    g = nx.DiGraph([(u, v) for (u, v) in g.edges() if u < v])
    return g


def nx_to_directed(graph: nx.Graph) -> nx.DiGraph:
    """Randomly chooses a direction for each edge."""
    if isinstance(graph, nx.DiGraph):
        return graph

    dedges = set()
    digraph = nx.DiGraph()

    for node, data in graph.nodes(data=True):
        digraph.add_node(node, **data)

    for u, v, data in graph.edges.data():
        edge = (u, v)
        inv_edge = (v, u)
        if edge in dedges or inv_edge in dedges:
            continue

        if np.random.default_rng().random() > 0.5:
            digraph.add_edge(*edge, **data)
            dedges.add(edge)
        else:
            digraph.add_edge(*inv_edge, **data)
            dedges.add(inv_edge)
    return digraph


def largest_component(graph: nx.Graph) -> nx.Graph:
    get_components = nx.weakly_connected_components \
        if isinstance(graph, nx.DiGraph) else nx.connected_components
    largest = max(get_components(graph), key=len)
    return graph.subgraph(largest)


def relabel_nx_graph(graph: nx.Graph, available_names: Collection[str]) -> nx.Graph:
    """Randomly label nodes with 'name' attribute in nx.Graph
    given list of available labels"""
    names = np.random.choice(available_names, size=graph.number_of_nodes())
    attributes = {node_id: {'name': name}
                  for node_id, name in zip(graph.nodes, names)}
    nx.set_node_attributes(graph, attributes)
    return graph


def postprocess_nx_graph(graph: nx.Graph,
                         node_labels: Optional[Sequence[str]] = ('x',),
                         connected: bool = True,
                         directed: bool = True):
    """Generate randomly labeled graph, enforce connectedness and direction."""
    # Remove unconnected components
    if connected:
        graph = largest_component(graph)
    # Optionally choose random directions for each edge
    if directed:
        # reverse() is mainly for trees
        # to make them growing towards root
        graph = nx_to_directed(graph).reverse()
    # Label the nodes randomly from available nodes
    if node_labels:
        graph = relabel_nx_graph(graph, node_labels)
    return graph


def generate_labeled_graph(kind: str,
                           size: int,
                           node_labels: Optional[Sequence[str]] = ('x',),
                           connected: bool = True,
                           directed: bool = True):
    """Generate randomly labeled graph of the specified kind and size,
    optionally enforce connectedness and direction. Important! With small specified size
    some methods can generate smaller graphs due to removal of unconnected components."""
    nx_graph = graph_generators[kind](size)
    graph = postprocess_nx_graph(nx_graph, node_labels, connected, directed)
    return graph


def _draw_sample_graphs(kind: str = 'gnp', sizes=tuple(range(5, 50, 5))):
    graphs = [generate_labeled_graph(kind, n) for n in sizes]
    draw_graphs_subplots(*graphs)


if __name__ == '__main__':
    _draw_sample_graphs('gnp')



from __future__ import annotations

import csv
import io
import itertools
import os
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union, TYPE_CHECKING

from golem.core.log import default_log
from golem.core.optimisers.objective.objective import ObjectiveInfo
from golem.core.optimisers.opt_history_objects.generation import Generation

from golem.core.paths import default_data_dir
from golem.serializers.serializer import default_load, default_save
from golem.visualisation.opt_viz import OptHistoryVisualizer

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.optimisers.opt_history_objects.individual import Individual


class OptHistory:
    """
    Contains optimization history, saves history to csv.
    Can be used for any type of graph that is serializable with Serializer.

    Args:
        objective: information about metrics (metric names and if it's multi-objective)
        default_save_dir: default directory used for saving history when not explicit path is provided.
    """

    def __init__(self,
                 objective: Optional[ObjectiveInfo] = None,
                 default_save_dir: Optional[os.PathLike] = None):
        self._objective = objective or ObjectiveInfo()
        self._generations: List[Generation] = []
        self.archive_history: List[List[Individual]] = []
        self._tuning_result: Optional[Graph] = None

        # init default save directory
        if default_save_dir:
            default_save_dir = Path(default_save_dir)
            if not default_save_dir.is_absolute():
                # if path is not absolute, treat it as relative to data dir
                default_save_dir = Path(default_data_dir()) / default_save_dir
        else:
            default_save_dir = default_data_dir()
        self._default_save_dir = str(default_save_dir)

    @property
    def objective(self):
        return self._objective

    def is_empty(self) -> bool:
        return not self.generations

    def add_to_history(self, individuals: Sequence[Individual], generation_label: Optional[str] = None,
                       generation_metadata: Optional[Dict[str, Any]] = None):
        generation = Generation(individuals, self.generations_count, generation_label, generation_metadata)
        self.generations.append(generation)

    def add_to_archive_history(self, individuals: Sequence[Individual]):
        self.archive_history.append(list(individuals))

    def to_csv(self, save_dir: Optional[os.PathLike] = None, file: os.PathLike = 'history.csv'):
        save_dir = save_dir or self._default_save_dir
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)

        with open(Path(save_dir, file), 'w', newline='') as file:
            writer = csv.writer(file, quoting=csv.QUOTE_ALL)

            # Write header
            metric_str = 'metric'
            if self.objective.is_multi_objective:
                metric_str += 's'
            header_row = ['index', 'generation', metric_str, 'quantity_of_operations', 'depth', 'metadata']
            writer.writerow(header_row)

            # Write history rows
            idx = 0
            for gen_num, gen_inds in enumerate(self.generations):
                for ind_num, ind in enumerate(gen_inds):
                    row = [idx, gen_num, ind.fitness.values, ind.graph.length, ind.graph.depth, ind.metadata]
                    writer.writerow(row)
                    idx += 1

    def save_current_results(self, save_dir: Optional[os.PathLike] = None):
        # Create folder if it's not exists
        save_dir = save_dir or self._default_save_dir
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)
            self._log.info(f"Created directory for saving optimization history: {save_dir}")

        try:
            last_gen_id = self.generations_count - 1
            last_gen = self.generations[last_gen_id]
            for individual in last_gen:
                ind_path = Path(save_dir, str(last_gen_id), str(individual.uid))
                ind_path.mkdir(exist_ok=True, parents=True)
                individual.save(json_file_path=ind_path / f'{str(individual.uid)}.json')
        except Exception as ex:
            self._log.exception(ex)

    def save(self, json_file_path: Union[str, os.PathLike] = None, is_save_light: bool = False) -> Optional[str]:
        """ Saves history to specified path.
        Args:
            json_file_path: path to json file where to save history.
            is_save_light: bool parameter to specify whether there is a need to save full history or a light version.
            NB! For experiments and etc. full histories must be saved. However, to make the analysis of results faster
            (show fitness plots, for example) the light version of histories can be saved too.
        """
        history_to_save = lighten_history(self) if is_save_light else self
        return default_save(obj=history_to_save, json_file_path=json_file_path)

    @staticmethod
    def load(json_str_or_file_path: Union[str, os.PathLike] = None) -> OptHistory:
        return default_load(json_str_or_file_path)

    @staticmethod
    def clean_results(dir_path: Optional[str] = None):
        """Clearn the directory tree with previously dumped history results."""
        if dir_path and os.path.isdir(dir_path):
            shutil.rmtree(dir_path, ignore_errors=True)
            os.mkdir(dir_path)

    @property
    def historical_fitness(self) -> Sequence[Sequence[Union[float, Sequence[float]]]]:
        """Return sequence of histories of generations per each metric"""
        if self.objective.is_multi_objective:
            historical_fitness = []
            num_metrics = len(self.generations[0][0].fitness.values)
            for objective_num in range(num_metrics):
                # history of specific objective for each generation
                objective_history = [[ind.fitness.values[objective_num] for ind in generation]
                                     for generation in self.generations]
                historical_fitness.append(objective_history)
        else:
            historical_fitness = [[ind.fitness.value for ind in pop]
                                  for pop in self.generations]
        return historical_fitness

    @property
    def all_historical_fitness(self) -> List[float]:
        historical_fitness = self.historical_fitness
        if self.objective.is_multi_objective:
            all_historical_fitness = []
            for obj_num in range(len(historical_fitness)):
                all_historical_fitness.append(list(itertools.chain(*historical_fitness[obj_num])))
        else:
            all_historical_fitness = list(itertools.chain(*historical_fitness))
        return all_historical_fitness

    def all_historical_quality(self, metric_position: int = 0) -> List[float]:
        """
        Return fitness history of population for specified metric.

        Args:
            metric_position: Index of the metric for multi-objective optimization.
             By default, choose first metric, assuming it is primary quality metric.

        Returns:
            List: all historical fitness
        """
        if self.objective.is_multi_objective:
            all_historical_quality = self.all_historical_fitness[metric_position]
        else:
            all_historical_quality = self.all_historical_fitness
        return all_historical_quality

    @property
    def show(self):
        return OptHistoryVisualizer(self)

    def get_leaderboard(self, top_n: int = 10) -> str:
        """
        Prints ordered description of the best solutions in history
        :param top_n: number of solutions to print
        """
        # Take only the first graph's appearance in history
        individuals_with_positions \
            = list({ind.graph.descriptive_id: (ind, gen_num, ind_num)
                    for gen_num, gen in enumerate(self.generations)
                    for ind_num, ind in reversed(list(enumerate(gen)))}.values())

        top_individuals = sorted(individuals_with_positions,
                                 key=lambda pos_ind: pos_ind[0].fitness, reverse=True)[:top_n]

        output = io.StringIO()
        separator = ' | '
        header = separator.join(['Position', 'Fitness', 'Generation', 'Graph'])
        print(header, file=output)
        for ind_num, ind_with_position in enumerate(top_individuals):
            individual, gen_num, ind_num = ind_with_position
            positional_id = f'g{gen_num}-i{ind_num}'
            print(separator.join([f'{ind_num:>3}, '
                                  f'{str(individual.fitness):>8}, '
                                  f'{positional_id:>8}, '
                                  f'{individual.graph.descriptive_id}']), file=output)

        # add info about initial assumptions (stored as zero generation)
        for i, individual in enumerate(self.generations[0]):
            ind = f'I{i}'
            positional_id = '-'
            print(separator.join([f'{ind:>3}'
                                  f'{str(individual.fitness):>8}',
                                  f'{positional_id}',
                                  f'{individual.graph.descriptive_id}']), file=output)
        return output.getvalue()

    @property
    def initial_assumptions(self) -> Optional[Generation]:
        if not self.generations:
            return None
        for gen in self.generations:
            if gen.label == 'initial_assumptions':
                return gen

    @property
    def final_choices(self) -> Optional[Generation]:
        if not self.generations:
            return None
        for gen in reversed(self.generations):
            if gen.label == 'final_choices':
                return gen

    @property
    def generations_count(self) -> int:
        return len(self.generations)

    @property
    def tuning_result(self):
        if hasattr(self, '_tuning_result'):
            return self._tuning_result
        else:
            return None

    @tuning_result.setter
    def tuning_result(self, val):
        self._tuning_result = val

    @property
    def generations(self):
        return self._generations

    @generations.setter
    def generations(self, value):
        self._generations = value

    @property
    def individuals(self):
        self._log.warning(
            '"OptHistory.individuals" is deprecated and will be removed later. '
            'Please, use "OptHistory.generations" to access generations.')
        return self.generations

    @individuals.setter
    def individuals(self, value):
        self.generations = value

    @property
    def _log(self):
        return default_log(self)


def lighten_history(history: OptHistory) -> OptHistory:
    """ Keeps the most informative field in OptHistory object to show most of the visualizations
    without excessive memory usage. """
    light_history = OptHistory()
    light_history._generations = \
        [Generation(iterable=gen, generation_num=i) for i, gen in enumerate(history.archive_history)]
    light_history.archive_history = history.archive_history
    light_history._objective = history.objective
    light_history._tuning_result = history.tuning_result
    return light_history



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Optional, Sequence, Union

from tqdm import tqdm

from golem.core.adapter import BaseOptimizationAdapter, IdentityAdapter
from golem.core.dag.graph import Graph
from golem.core.dag.graph_verifier import GraphVerifier, VerifierRuleType
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.log import default_log
from golem.core.optimisers.advisor import DefaultChangeAdvisor
from golem.core.optimisers.optimization_parameters import OptimizationParameters
from golem.core.optimisers.genetic.evaluation import DelegateEvaluator
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import GraphFunction, Objective, ObjectiveFunction
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.core.optimisers.opt_node_factory import DefaultOptNodeFactory, OptNodeFactory
from golem.core.optimisers.random_graph_factory import RandomGraphFactory, RandomGrowthGraphFactory
from golem.utilities.random import RandomStateHandler

STRUCTURAL_DIVERSITY_FREQUENCY_CHECK = 5


def do_nothing_callback(*args, **kwargs):
    pass


@dataclass
class AlgorithmParameters:
    """Base class for definition of optimizers-specific parameters.
    Can be extended for custom optimizers.

    :param multi_objective: defines if the optimizer must be multi-criterial
    :param offspring_rate: offspring rate used on next population
    :param pop_size: initial population size
    :param max_pop_size: maximum population size; optional, if unspecified, then population size is unbound
    :param adaptive_depth: flag to enable adaptive configuration of graph depth
    :param adaptive_depth_max_stagnation: max number of stagnating populations before adaptive depth increment
    """

    multi_objective: bool = False
    offspring_rate: float = 0.5
    pop_size: int = 20
    max_pop_size: Optional[int] = 55
    adaptive_depth: bool = False
    adaptive_depth_max_stagnation: int = 3
    structural_diversity_frequency_check: int = STRUCTURAL_DIVERSITY_FREQUENCY_CHECK


@dataclass
class GraphGenerationParams:
    """
    This dataclass is for defining the parameters using in graph generation process

    :param adapter: instance of domain graph adapter for adaptation
     between domain and optimization graphs
    :param rules_for_constraint: collection of constraints for graph verification
    :param advisor: instance providing task and context-specific advices for graph changes
    :param node_factory: instance for generating new nodes in the process of graph search
    :param remote_evaluator: instance of delegate evaluator for evaluation of graphs
    """
    adapter: BaseOptimizationAdapter
    verifier: GraphVerifier
    advisor: DefaultChangeAdvisor
    node_factory: OptNodeFactory
    random_graph_factory: RandomGraphFactory
    remote_evaluator: Optional[DelegateEvaluator] = None

    def __init__(self, adapter: Optional[BaseOptimizationAdapter] = None,
                 rules_for_constraint: Sequence[VerifierRuleType] = tuple(DEFAULT_DAG_RULES),
                 advisor: Optional[DefaultChangeAdvisor] = None,
                 node_factory: Optional[OptNodeFactory] = None,
                 random_graph_factory: Optional[RandomGraphFactory] = None,
                 available_node_types: Optional[Sequence[Any]] = None,
                 remote_evaluator: Optional[DelegateEvaluator] = None,
                 ):
        self.adapter = adapter or IdentityAdapter()
        self.verifier = GraphVerifier(rules_for_constraint, self.adapter)
        self.advisor = advisor or DefaultChangeAdvisor()
        self.remote_evaluator = remote_evaluator
        if node_factory:
            self.node_factory = node_factory
        elif available_node_types:
            self.node_factory = DefaultOptNodeFactory(available_node_types)
        else:
            self.node_factory = DefaultOptNodeFactory()
        self.random_graph_factory = random_graph_factory or RandomGrowthGraphFactory(self.verifier,
                                                                                     self.node_factory)


class GraphOptimizer:
    """
    Base class of graph optimizer. It allows to find the optimal solution using specified metric (one or several).
    To implement the specific optimisation method,
    the abstract method 'optimize' should be re-defined in the ancestor class
    (e.g.  PopulationalOptimizer, RandomSearchGraphOptimiser, etc).

    :param objective: objective for optimisation
    :param initial_graphs: graphs which were initialized outside the optimizer
    :param requirements: implementation-independent requirements for graph optimizer
    :param graph_generation_params: parameters for new graph generation
    :param graph_optimizer_params: parameters for specific implementation of graph optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Optional[Sequence[Union[Graph, Any]]] = None,
                 # TODO: rename params to avoid confusion
                 requirements: Optional[OptimizationParameters] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_params: Optional[AlgorithmParameters] = None):
        self.log = default_log(self)
        self._objective = objective
        initial_graphs = graph_generation_params.adapter.adapt(initial_graphs) if initial_graphs else None
        self.initial_graphs = [graph for graph in initial_graphs if graph_generation_params.verifier(graph)] \
            if initial_graphs else None
        self.requirements = requirements or OptimizationParameters()
        self.graph_generation_params = graph_generation_params or GraphGenerationParams()
        self.graph_optimizer_params = graph_optimizer_params or AlgorithmParameters()
        self._iteration_callback: IterationCallback = do_nothing_callback
        self._history = OptHistory(objective.get_info(), requirements.history_dir) \
            if requirements and requirements.keep_history else None
        # Log random state for reproducibility of runs
        RandomStateHandler.log_random_state()

    @property
    def objective(self) -> Objective:
        """Returns Objective of this optimizer with information about metrics used."""
        return self._objective

    @property
    def history(self) -> Optional[OptHistory]:
        """Returns optimization history"""
        return self._history

    @abstractmethod
    def optimise(self, objective: ObjectiveFunction) -> Sequence[OptGraph]:
        """
        Method for running of optimization using specified algorithm.
        :param objective: objective function that specifies optimization target
        :return: sequence of the best graphs
        """
        pass

    def set_iteration_callback(self, callback: Optional[Callable]):
        """Set optimisation callback that is called at the end of
        each iteration, with the next generation passed as argument.
        Resets the callback if None is passed."""
        self._iteration_callback = callback or do_nothing_callback

    def set_evaluation_callback(self, callback: Optional[GraphFunction]):
        """Set or reset (with None) post-evaluation callback
        that's called on each graph after its evaluation."""
        pass

    @property
    def _progressbar(self):
        if self.requirements.show_progress:
            bar = tqdm(total=self.requirements.num_of_generations, desc='Generations', unit='gen', initial=0)
        else:
            # disable call to tqdm.__init__ to avoid stdout/stderr access inside it
            # part of a workaround for https://github.com/nccr-itmo/FEDOT/issues/765
            bar = EmptyProgressBar()
        return bar


IterationCallback = Callable[[PopulationT, GraphOptimizer], Any]


class EmptyProgressBar:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True

    def close(self):
        return

    def update(self):
        return



import random
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union, Sequence, Optional

import numpy as np

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.log import default_log
from golem.core.optimisers.adaptive.common_types import ObsType, ActType
from golem.core.optimisers.adaptive.experience_buffer import ExperienceBuffer


class MutationAgentTypeEnum(Enum):
    default = 'default'
    random = 'random'
    bandit = 'bandit'
    contextual_bandit = 'contextual_bandit'
    neural_bandit = 'neural_bandit'


class OperatorAgent(ABC):
    def __init__(self, actions: Sequence[ActType], enable_logging: bool = True):
        self.actions = list(actions)
        self._enable_logging = enable_logging
        self._log = default_log(self)

    @property
    def available_actions(self) -> Sequence[ActType]:
        return self.actions

    @abstractmethod
    def partial_fit(self, experience: ExperienceBuffer):
        raise NotImplementedError()

    @abstractmethod
    def choose_action(self, obs: Optional[ObsType]) -> ActType:
        raise NotImplementedError()

    @abstractmethod
    def choose_nodes(self, graph: ObsType, num_nodes: int = 1) -> Union[GraphNode, Sequence[GraphNode]]:
        raise NotImplementedError()

    @abstractmethod
    def get_action_probs(self, obs: Optional[ObsType]) -> Sequence[float]:
        raise NotImplementedError()

    @abstractmethod
    def get_action_values(self, obs: Optional[ObsType]) -> Sequence[float]:
        raise NotImplementedError()

    def _dbg_log(self, obs, actions, rewards):
        if self._enable_logging:
            prec = 4
            rr = np.array(rewards).round(prec)
            nonzero = rr[rr.nonzero()]
            msg = f'len={len(rr)} nonzero={len(nonzero)} '
            if len(nonzero) > 0:
                msg += (f'avg={nonzero.mean():.3f} std={nonzero.std():.3f} '
                        f'min={nonzero.min():.3f} max={nonzero.max():.3f} ')

            self._log.info(msg)
            self._log.info(f'actions/rewards: {list(zip(actions, rr))}')

            action_values = list(map(self.get_action_values, obs))
            action_probs = list(map(self.get_action_probs, obs))
            action_values = np.round(np.mean(action_values, axis=0), prec)
            action_probs = np.round(np.mean(action_probs, axis=0), prec)

            self._log.info(f'exp={action_values} '
                           f'probs={action_probs}')


class RandomAgent(OperatorAgent):
    def __init__(self,
                 actions: Sequence[ActType],
                 probs: Optional[Sequence[float]] = None,
                 enable_logging: bool = True):
        super().__init__(actions, enable_logging)
        self._probs = probs or [1. / len(actions)] * len(actions)

    def choose_action(self, obs: Graph) -> ActType:
        action = np.random.choice(self.actions, p=self.get_action_probs(obs))
        return action

    def choose_nodes(self, graph: Graph, num_nodes: int = 1) -> Union[GraphNode, Sequence[GraphNode]]:
        subject_nodes = random.sample(graph.nodes, k=num_nodes)
        return subject_nodes[0] if num_nodes == 1 else subject_nodes

    def partial_fit(self, experience: ExperienceBuffer):
        obs, actions, rewards = experience.retrieve_experience()
        self._dbg_log(obs, actions, rewards)

    def get_action_probs(self, obs: Optional[Graph] = None) -> Sequence[float]:
        return self._probs

    def get_action_values(self, obs: Optional[Graph] = None) -> Sequence[float]:
        return self._probs



from typing import Tuple, Dict, TYPE_CHECKING
from uuid import uuid4

import networkx as nx

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.dag.graph_node import GraphNode


def graph_structure_as_nx_graph(structural_graph: 'Graph') -> Tuple[nx.DiGraph, Dict[uuid4, 'GraphNode']]:
    """ Convert graph into networkx graph object """
    nx_graph = nx.DiGraph()
    node_labels = {}
    new_node_indices = {}
    for node in structural_graph.nodes:
        unique_id = uuid4()
        node_labels[unique_id] = node
        new_node_indices[node] = unique_id
        nx_graph.add_node(unique_id)

    def add_edges(nx_graph, structural_graph, new_node_indices):
        for node in structural_graph.nodes:
            for parent in node.nodes_from:
                nx_graph.add_edge(new_node_indices[parent],
                                  new_node_indices[node])

    add_edges(nx_graph, structural_graph, new_node_indices)
    return nx_graph, node_labels



from datetime import timedelta
from functools import partial
from typing import Optional, Sequence, Callable, List
import networkx as nx

from examples.adaptive_optimizer.experiment_setup import run_adaptive_mutations
from examples.synthetic_graph_evolution.graph_search import graph_search_setup
from examples.synthetic_graph_evolution.generators import postprocess_nx_graph
from examples.synthetic_graph_evolution.tree_search import tree_search_setup
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.metrics.edit_distance import tree_edit_dist
from golem.metrics.graph_metrics import spectral_dist, size_diff, degree_distance


def generate_gnp_graphs(graph_size: int,
                        gnp_probs: Sequence[float] = (0.05, 0.1, 0.15, 0.3),
                        node_types: Sequence[str] = ('x',)
                        ):
    targets = []
    for prob in gnp_probs:
        # Generate target graph
        nx_graph = nx.gnp_random_graph(graph_size, prob, directed=True)
        target_graph = postprocess_nx_graph(nx_graph, node_labels=node_types)
        targets.append(target_graph)
    return targets


def generate_trees(graph_sizes: Sequence[int], node_types: Sequence[str] = ('x',)):
    trees = [nx.random_tree(n, create_using=nx.DiGraph) for n in graph_sizes]
    trees = [postprocess_nx_graph(g, node_labels=node_types) for g in trees]
    return trees


def get_graph_gp_params(objective: Objective, adaptive_mutation_type: MutationAgentTypeEnum,
                        context_agent_type: ContextAgentTypeEnum = None,
                        mutation_types: List[MutationTypesEnum] = None, pop_size: int = None,
                        decaying_factor: float = 1.0, window_size: int = 5):
    mutation_types = mutation_types or [MutationTypesEnum.single_change,
                                        MutationTypesEnum.single_add,
                                        MutationTypesEnum.single_drop]
    return GPAlgorithmParameters(
        adaptive_mutation_type=adaptive_mutation_type,
        context_agent_type=context_agent_type,
        pop_size=pop_size or 21,
        multi_objective=objective.is_multi_objective,
        genetic_scheme_type=GeneticSchemeTypesEnum.generational,
        mutation_types=mutation_types,
        crossover_types=[CrossoverTypesEnum.none],
        decaying_factor=decaying_factor,
        window_size=window_size
    )


def run_experiment_node_num(adaptive_mutation_type: MutationAgentTypeEnum,
                            context_agent_type: Optional[ContextAgentTypeEnum] = None,
                            target_sizes: Sequence[int] = (50, 400),
                            trial_timeout: int = 15,
                            run_func: Callable = run_adaptive_mutations):
    for target_size in target_sizes:
        # Setup simple objective that searches for required graph size (number of nodes)
        objective = Objective({'graph_size': lambda graph: abs(target_size -
                                                               graph.number_of_nodes())})

        # Build the optimizer
        optimizer, _ = graph_search_setup(
            objective=objective,
            optimizer_cls=EvoGraphOptimizer,
            algorithm_parameters=get_graph_gp_params(objective=objective,
                                                     adaptive_mutation_type=adaptive_mutation_type,
                                                     context_agent_type=context_agent_type),
            timeout=timedelta(minutes=trial_timeout),
            num_iterations=target_size * 3
        )
        run_func(optimizer, objective, visualize=True)


def run_experiment_edge_num(adaptive_mutation_type: MutationAgentTypeEnum,
                            context_agent_type: Optional[ContextAgentTypeEnum] = None,
                            target_sizes: Sequence[int] = (100, 400),
                            trial_timeout: int = 15,
                            run_func: Callable = run_adaptive_mutations):
    for target_size in target_sizes:
        # Setup simple objective that searches for required graph size (number of nodes)
        objective = Objective({'graph_size': lambda graph: abs(target_size -
                                                               graph.number_of_edges())})

        # Build the optimizer
        optimizer, _ = graph_search_setup(
            objective=objective,
            optimizer_cls=EvoGraphOptimizer,
            algorithm_parameters=get_graph_gp_params(objective=objective,
                                                     adaptive_mutation_type=adaptive_mutation_type,
                                                     context_agent_type=context_agent_type),
            timeout=timedelta(minutes=trial_timeout),
            num_iterations=target_size * 3,
        )
        run_func(optimizer, objective, visualize=True)


def run_experiment_graphs_ratio_edges_nodes(adaptive_mutation_type: MutationAgentTypeEnum,
                                            context_agent_type: Optional[ContextAgentTypeEnum] = None,
                                            trial_timeout: int = 15,
                                            trial_iterations: Optional[int] = 500,
                                            run_func: Callable = run_adaptive_mutations):
    """In this experiment setup we generate different graphs with different ratios of #Edges/#Nodes.
    Respectively, probabilities of adding edges and adding nodes must be different for different targets."""

    node_types = ['x']
    for target in generate_gnp_graphs(gnp_probs=[0.15, 0.3], graph_size=100, node_types=node_types):
        # Setup objective that measures some graph-theoretic similarity measure
        objective = Objective(
            quality_metrics={
                'sp_adj': partial(spectral_dist, target, kind='adjacency'),
                'sp_lapl': partial(spectral_dist, target, kind='laplacian'),
            },
            complexity_metrics={
                'graph_size': partial(size_diff, target),
                'degree': partial(degree_distance, target),
            },
            is_multi_objective=True,
        )

        # Build the optimizer
        optimizer, _ = graph_search_setup(
            objective=objective,
            optimizer_cls=EvoGraphOptimizer,
            algorithm_parameters=get_graph_gp_params(objective=objective,
                                                     adaptive_mutation_type=adaptive_mutation_type,
                                                     context_agent_type=context_agent_type),
            node_types=node_types,
            timeout=timedelta(minutes=trial_timeout),
            num_iterations=trial_iterations,
        )

        run_func(optimizer, objective, target, visualize=True)


def run_experiment_trees(adaptive_mutation_type: MutationAgentTypeEnum,
                         context_agent_type: Optional[ContextAgentTypeEnum] = None,
                         trial_timeout: int = 15,
                         trial_iterations: Optional[int] = 500,
                         run_func: Callable = run_adaptive_mutations):
    node_types = ['x']
    for target in generate_trees(graph_sizes=[20, 30, 50], node_types=node_types):
        # Setup objective that measures some graph-theoretic similarity measure
        objective = Objective(
            quality_metrics={'edit_dist': partial(tree_edit_dist, target)},
            complexity_metrics={'degree': partial(degree_distance, target)},
            is_multi_objective=False,
        )

        gp_params = GPAlgorithmParameters(
            multi_objective=objective.is_multi_objective,
            mutation_types=[
                MutationTypesEnum.single_add,
                MutationTypesEnum.single_drop,
            ],
            crossover_types=[CrossoverTypesEnum.none],
            adaptive_mutation_type=adaptive_mutation_type,
            context_agent_type=context_agent_type
        )

        # Build the optimizer
        optimizer, _ = tree_search_setup(
            objective=objective,
            optimizer_cls=EvoGraphOptimizer,
            algorithm_parameters=gp_params,
            node_types=node_types,
            timeout=timedelta(minutes=trial_timeout),
            num_iterations=trial_iterations,
        )

        run_func(optimizer, objective, target, visualize=True)


if __name__ == '__main__':
    """Run adaptive optimizer on different targets to see how adaptive agent converges
    to different probabilities of actions (i.e. mutations) for different targets."""
    adaptive_mutation_type = MutationAgentTypeEnum.bandit

    run_experiment_node_num(trial_timeout=2, adaptive_mutation_type=adaptive_mutation_type)
    run_experiment_edge_num(trial_timeout=2, adaptive_mutation_type=adaptive_mutation_type)
    run_experiment_trees(trial_timeout=10, trial_iterations=2000, adaptive_mutation_type=adaptive_mutation_type)
    run_experiment_graphs_ratio_edges_nodes(trial_timeout=10, trial_iterations=2000,
                                            adaptive_mutation_type=adaptive_mutation_type)



from golem.core.log import default_log
from golem.core.dag.graph import Graph
from golem.structural_analysis.graph_sa.results.utils import get_entity_from_str


def nodes_deletion(graph: Graph, worst_result: dict) -> Graph:
    """ Extracts the node index from the entity key and removes it from the graph """

    node_to_delete = graph.nodes[int(worst_result["entity_idx"])]

    graph.delete_node(node_to_delete)
    default_log('NodeDeletion').message(f'{node_to_delete.name} was deleted')

    return graph


def nodes_replacement(graph: Graph, worst_result: dict) -> Graph:
    """ Extracts the node index and the operation to which it needs to be replaced from the entity key
    and replaces the node with a new one """

    # get the node that will be replaced
    node_to_replace = get_entity_from_str(graph=graph, entity_str=worst_result["entity_idx"])
    # get node to replace to
    new_node = graph.nodes[0].__class__(worst_result["entity_to_replace_to"])
    # new_node = graph.nodes[int(worst_result["entity_to_replace_to"])]
    new_node.nodes_from = []

    graph.update_node(old_node=node_to_replace, new_node=new_node)

    default_log('NodeReplacement').message(f'{node_to_replace.name} was replaced with {new_node.name}')

    return graph


def subtree_deletion(graph: Graph, worst_result: dict) -> Graph:
    """ Extracts the node index from the entity key and removes its subtree from the graph """

    node_to_delete = get_entity_from_str(graph=graph, entity_str=worst_result["entity_idx"])
    graph.delete_subtree(node_to_delete)
    default_log('SubtreeDeletion').message(f'{node_to_delete.name} subtree was deleted')

    return graph


def edges_deletion(graph: Graph, worst_result: dict) -> Graph:
    """ Extracts the edge's nodes indices from the entity key and removes edge from the graph """
    parent_node, child_node = get_entity_from_str(graph=graph, entity_str=worst_result["entity_idx"])

    graph.disconnect_nodes(parent_node, child_node)
    default_log('EdgeDeletion').message(f'Edge from {parent_node.name} to {child_node.name} was deleted')

    return graph


def edges_replacement(graph: Graph, worst_result: dict) -> Graph:
    """ Extracts the edge's nodes indices and the new edge to which it needs to be replaced from the entity key
    and replaces the edge with a new one """

    # get the edge that will be replaced
    parent_node, child_node = get_entity_from_str(graph=graph, entity_str=worst_result["entity_idx"])

    # get an edge to replace
    next_parent_node, next_child_node = \
        get_entity_from_str(graph=graph, entity_str=worst_result["entity_to_replace_to"])
    graph.connect_nodes(next_parent_node, next_child_node)

    graph.disconnect_nodes(parent_node, child_node, clean_up_leftovers=False)

    default_log('EdgeReplacement').message(f'Edge from {parent_node.name} to {child_node.name} was replaced with '
                                           f'edge from {next_parent_node.name} to {next_child_node.name}')

    return graph



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Hashable, Sequence
from uuid import uuid4

from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.utilities.data_structures import ensure_wrapped_in_sequence


@dataclass(frozen=True)
class ParentOperator:
    type_: str
    operators: Sequence[Hashable]
    parent_individuals: Sequence[Individual] = field()
    uid: str = field(default_factory=lambda: str(uuid4()), init=False)

    def __post_init__(self):
        operators = ensure_wrapped_in_sequence(self.operators)
        object.__setattr__(self, 'operators', tuple(operators))
        parent_individuals = ensure_wrapped_in_sequence(self.parent_individuals)
        object.__setattr__(self, 'parent_individuals', tuple(parent_individuals))

    def __repr__(self):
        return (f'<ParentOperator {self.uid} | type: {self.type_} | operators: {self.operators} '
                f'| parent_individuals({len(self.parent_individuals)}): {self.parent_individuals}>')



import random
from abc import ABC
from copy import deepcopy
from os import makedirs
from os.path import exists, join
from typing import List, Optional, Type, Union, Any

from golem.core.log import default_log
from golem.core.dag.graph import Graph, GraphNode
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_node_factory import OptNodeFactory
from golem.core.optimisers.timer import OptimisationTimer
from golem.core.paths import default_data_dir
from golem.structural_analysis.base_sa_approaches import BaseAnalyzeApproach
from golem.structural_analysis.graph_sa.results.deletion_sa_approach_result import \
    DeletionSAApproachResult
from golem.structural_analysis.graph_sa.results.object_sa_result import ObjectSAResult
from golem.structural_analysis.graph_sa.results.replace_sa_approach_result import \
    ReplaceSAApproachResult
from golem.structural_analysis.graph_sa.results.utils import EntityTypesEnum
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements, \
    ReplacementAnalysisMetaParams


class NodeAnalysis:
    """
    :param approaches: methods applied to nodes to modify the graph or analyze certain operations.\
    Default: [NodeDeletionAnalyze, NodeTuneAnalyze, NodeReplaceOperationAnalyze]
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, node_factory: Any,
                 approaches: Optional[List[Type['NodeAnalyzeApproach']]] = None,
                 approaches_requirements: Optional[StructuralAnalysisRequirements] = None,
                 path_to_save: Optional[str] = None):

        self.node_factory = node_factory

        self.approaches = [NodeDeletionAnalyze, NodeReplaceOperationAnalyze] if approaches is None else approaches

        self.path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural') if path_to_save is None else path_to_save
        self.log = default_log(self)

        self.approaches_requirements = \
            StructuralAnalysisRequirements() if approaches_requirements is None else approaches_requirements

    def analyze(self, graph: Graph, node: GraphNode,
                objective: Objective,
                timer: Optional[OptimisationTimer] = None) -> ObjectSAResult:

        """
        Method runs Node analysis within defined approaches

        :param graph: Graph containing the analyzed Node
        :param node: Node object to analyze in Graph
        :param objective: objective function for computing metric values
        :param timer: timer to check if the time allotted for structural analysis has expired
        :return: dict with Node analysis result per approach
        """

        results = ObjectSAResult(entity_idx=str(graph.nodes.index(node)),
                                 entity_type=EntityTypesEnum.node)

        for approach in self.approaches:
            if timer is not None and timer.is_time_limit_reached():
                break

            results.add_result(approach(graph=graph,
                                        objective=objective,
                                        node_factory=self.node_factory,
                                        requirements=self.approaches_requirements,
                                        path_to_save=self.path_to_save).analyze(node=node))
        return results


class NodeAnalyzeApproach(BaseAnalyzeApproach, ABC):
    """
    Base class for node analysis approach.
    :param graph: Graph containing the analyzed Node
    :param objective: objective functions for computing metric values
    :param node_factory: node factory to choose nodes to replace to
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, graph: Graph, objective: Objective, node_factory: OptNodeFactory,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, requirements)
        self._node_factory = node_factory

        self._origin_metrics = list()
        self._path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural') if path_to_save is None else path_to_save
        self.log = default_log(prefix='node_analysis')

        if not exists(self._path_to_save):
            makedirs(self._path_to_save)


class NodeDeletionAnalyze(NodeAnalyzeApproach):
    def __init__(self, graph: Graph, objective: Objective,
                 node_factory: OptNodeFactory,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, node_factory, requirements)
        self._path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural') if path_to_save is None else path_to_save
        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    def analyze(self, node: GraphNode, **kwargs) -> DeletionSAApproachResult:
        """
        Receives a graph without the specified node and tries to calculate the loss for it

        :param node: GraphNode object to analyze
        :return: the ratio of modified graph score to origin score
        """
        results = DeletionSAApproachResult()
        if node is self._graph.root_node:
            self.log.warning(f'{node} node can not be deleted')
            results.add_results(metrics_values=[-1.0] * len(self._objective.metrics))
            return results
        else:
            shortened_graph = self.sample(node)
            if shortened_graph:
                losses = self._compare_with_origin_by_metrics(shortened_graph)
                self.log.message(f'losses for {node.name}: {losses}')
                del shortened_graph
            else:
                losses = [-1.0] * len(self._objective.metrics)

            results.add_results(metrics_values=losses)
            return results

    def sample(self, node: GraphNode):
        """
        Checks if it is possible to delete the node from the graph so that it remains valid,
        and if so, deletes

        :param node: GraphNode object to delete from Graph object
        :return: Graph object without node
        """
        graph_sample = deepcopy(self._graph)
        node_index_to_delete = self._graph.nodes.index(node)
        node_to_delete = graph_sample.nodes[node_index_to_delete]

        if node_to_delete.name == 'class_decompose':
            for child in graph_sample.node_children(node_to_delete):
                graph_sample.delete_node(child)

        graph_sample.delete_node(node_to_delete)

        verifier = self._requirements.graph_verifier
        if not verifier.verify(graph_sample):
            self.log.message('Can not delete node since modified graph can not be verified')
            return None

        return graph_sample


class NodeReplaceOperationAnalyze(NodeAnalyzeApproach):
    """
    Replace node with operations available for the current task
    and evaluate the score difference
    """

    def __init__(self, graph: Graph, objective: Objective,
                 node_factory: OptNodeFactory,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, node_factory, requirements)

        self._path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural') if path_to_save is None else path_to_save
        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    def analyze(self, node: GraphNode, **kwargs) -> ReplaceSAApproachResult:
        """
        Counts the loss on each changed graph received and returns losses

        :param node: GraphNode object to analyze

        :return: the ratio of modified graph score to origin score
        """
        result = ReplaceSAApproachResult()
        requirements: ReplacementAnalysisMetaParams = self._requirements.replacement_meta
        node_id = self._graph.nodes.index(node)
        samples = self.sample(node=node,
                              nodes_to_replace_to=requirements.nodes_to_replace_to,
                              number_of_random_operations=requirements.number_of_random_operations_nodes)

        for sample_graph in samples:
            loss_per_sample = self._compare_with_origin_by_metrics(sample_graph)
            self.log.message(f'losses: {loss_per_sample}\n')

            result.add_results(entity_to_replace_to=sample_graph.nodes[node_id].name, metrics_values=loss_per_sample)

        return result

    def sample(self, node: GraphNode,
               nodes_to_replace_to: Optional[List[GraphNode]],
               number_of_random_operations: int = 1) -> Union[List[Graph], Graph]:
        """
        Replaces the given node with a pool of nodes available for replacement (see _node_generation docstring)

        :param node: GraphNode object to replace
        :param nodes_to_replace_to: nodes provided for old_node replacement
        :param number_of_random_operations: number of replacement operations, \
        if nodes_to_replace_to not provided
        :return: Sequence of Graph objects with new operations instead of old one
        """

        if not nodes_to_replace_to:
            nodes_to_replace_to = self._node_generation(node=node,
                                                        node_factory=self._node_factory,
                                                        number_of_operations=number_of_random_operations)

        samples = list()
        for replacing_node in nodes_to_replace_to:
            sample_graph = deepcopy(self._graph)
            replaced_node_index = self._graph.nodes.index(node)
            replaced_node = sample_graph.nodes[replaced_node_index]
            sample_graph.update_node(old_node=replaced_node,
                                     new_node=replacing_node)
            verifier = self._requirements.graph_verifier
            if not verifier.verify(sample_graph):
                self.log.warning(f'Can not replace {node.name} node with {replacing_node.name} node.')
            else:
                self.log.message(f'replacing node: {replacing_node.name}')
                samples.append(sample_graph)

        if not samples:
            samples.append(self._graph)

        return samples

    @staticmethod
    def _node_generation(node: GraphNode,
                         node_factory: OptNodeFactory,
                         number_of_operations: int = 1) -> List[GraphNode]:
        """
        The method returns possible nodes that can replace the given node

        :param node: the node to be replaced
        :param number_of_operations: limits the number of possible nodes to replace to

        :return: nodes that can be used to replace
        """

        available_nodes = []
        for i in range(number_of_operations):
            available_nodes.append(node_factory.exchange_node(node=node))

        if number_of_operations:
            available_nodes = [i for i in available_nodes if i != node.name]
            number_of_operations = min(len(available_nodes), number_of_operations)
            random_nodes = random.sample(available_nodes, number_of_operations)
        else:
            random_nodes = available_nodes

        return random_nodes


class SubtreeDeletionAnalyze(NodeAnalyzeApproach):
    """
    Approach to delete specified node subtree
    """

    def __init__(self, graph: Graph, objective: Objective,
                 node_factory: OptNodeFactory,
                 requirements: StructuralAnalysisRequirements = None, path_to_save=None):
        super().__init__(graph, objective, node_factory, requirements)
        self._path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural')\
            if path_to_save is None else path_to_save
        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    def analyze(self, node: GraphNode, **kwargs) -> DeletionSAApproachResult:
        """
        Receives a graph without the specified node's subtree and
        tries to calculate the loss for it

        :param node: GraphNode object to analyze
        :return: the ratio of modified graph score to origin score
        """
        results = DeletionSAApproachResult()
        if node is self._graph.root_node:
            self.log.warning(f'{node} subtree can not be deleted')
            results.add_results(metrics_values=[-1.0] * len(self._objective.metrics))
            return results
        else:
            shortened_graph = self.sample(node)
            if shortened_graph:
                losses = self._compare_with_origin_by_metrics(shortened_graph)
                self.log.message(f'losses for {node.name}: {losses}')
                del shortened_graph
            else:
                losses = [-1.0] * len(self._objective.metrics)

            results.add_results(metrics_values=losses)
            return results

    def sample(self, node: GraphNode):
        """
        Checks if it is possible to delete the node's subtree from the graph so that it remains valid,
        and if so, deletes

        :param node: GraphNode object from which to delete subtree from Graph object
        :return: Graph object without subtree
        """
        graph_sample = deepcopy(self._graph)
        node_index_to_delete = self._graph.nodes.index(node)
        node_to_delete = graph_sample.nodes[node_index_to_delete]

        if node_to_delete.name == 'class_decompose':
            for child in graph_sample.node_children(node_to_delete):
                graph_sample.delete_node(child)

        graph_sample.delete_subtree(node_to_delete)

        verifier = self._requirements.graph_verifier
        if not verifier.verify(graph_sample):
            self.log.warning('Can not delete subtree since modified graph can not pass verification')
            return None

        return graph_sample



from typing import Optional, List, Callable, Tuple, Iterable

from golem.core.log import default_log

ConditionType = Callable[[], bool]
ConditionEntryType = Tuple[ConditionType, Optional[str]]


class GroupedCondition:
    """Represents sequence of ordinary conditions with logging.
    All composed conditions are combined with reduce function on booleans.

    By the default 'any' is used, so in this case the grouped condition is True
    if any of the composed conditions is True. The message corresponding
    to the actual fired condition is logged (if it was provided)."""

    def __init__(self, conditions_reduce: Callable[[Iterable[bool]], bool] = any, results_as_message: bool = False):
        self._reduce = conditions_reduce
        self._conditions: List[ConditionEntryType] = []
        self._log = default_log(self)
        self._results_as_message = results_as_message

    def add_condition(self, condition: ConditionType, log_msg: Optional[str] = None) -> 'GroupedCondition':
        """Builder-like method for adding conditions."""
        self._conditions.append((condition, log_msg))
        return self

    def __bool__(self):
        return self()

    def __call__(self) -> bool:
        return self._reduce(map(self._check_condition, self._conditions))

    def _check_condition(self, entry: ConditionEntryType) -> bool:
        cond, msg = entry
        res = cond()
        if res and msg:
            if self._results_as_message:
                self._log.message(msg)
            else:
                self._log.info(msg)
        return res



from abc import ABC, abstractmethod

from typing import Tuple, Union, Optional


class Serializable(ABC):

    @classmethod
    def from_serialized(cls, source: Union[str, dict], internal_state_data: Optional[dict] = None):
        """
        Static constructor for convenience. Creates default instance and calls .load() on it.
        """
        default_instance = cls()
        default_instance.load(source, internal_state_data)
        return default_instance

    @abstractmethod
    def save(self, path: str = None, datetime_in_path: bool = True) -> Tuple[str, dict]:
        """
        Save the graph to a json representation
        with pickled custom data (e.g. fitted models).

        :param path: path to json file with operation
        :param datetime_in_path: flag for addition of the datetime stamp to saving path
        :return: json containing a composite operation description
        """
        raise NotImplementedError()

    @abstractmethod
    def load(self, source: Union[str, dict], internal_state_data: Optional[dict] = None):
        """
        Load the graph from json representation, optionally
        with pickled custom internal data (e.g. fitted models).

        :param source: path to json file with operation or json dictionary itself
        :param internal_state_data: dictionary of the internal state
        """
        raise NotImplementedError()



from abc import ABC, abstractmethod
from copy import copy
from typing import List, Optional, Iterable
from uuid import uuid4


class GraphNode(ABC):
    """Definition of the node in directed graph structure.

    Provides interface for getting and modifying the parent nodes
    and recursive description based on all preceding nodes.
    """
    def __init__(self):
        self.uid = str(uuid4())

    @property
    @abstractmethod
    def nodes_from(self) -> List['GraphNode']:
        """Gets all parent nodes of this graph node

        Returns:
            List['GraphNode']: all the parent nodes
        """
        pass

    @nodes_from.setter
    @abstractmethod
    def nodes_from(self, nodes: Optional[Iterable['GraphNode']]):
        """Changes value of parent nodes of this graph node

        Args:
            nodes: new sequence of parent nodes
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """ Str name of this graph node """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Returns short node type description

        Returns:
            str: text graph node representation
        """
        pass

    def __repr__(self) -> str:
        """Returns full node description

        Returns:
            str: text graph node representation
        """
        return self.__str__()

    def description(self) -> str:
        """Returns full node description
        for use in recursive id.

        Returns:
            str: text graph node representation
        """

        return self.__str__()

    @property
    def descriptive_id(self) -> str:
        """Returns structural identifier of the subgraph starting at this node

        Returns:
            str: text description of the content in the node and its parameters
        """
        return descriptive_id_recursive(self)


def descriptive_id_recursive(current_node: GraphNode, visited_nodes=None) -> str:
    """ Returns descriptive id with nodes names. """
    if visited_nodes is None:
        visited_nodes = []

    node_label = current_node.description()

    full_path_items = []
    if current_node in visited_nodes:
        return 'ID_CYCLED'
    visited_nodes.append(current_node)
    if current_node.nodes_from:
        previous_items = []
        for parent_node in current_node.nodes_from:
            previous_items.append(f'{descriptive_id_recursive(parent_node, copy(visited_nodes))};')
        previous_items.sort()
        previous_items_str = ';'.join(previous_items)

        full_path_items.append(f'({previous_items_str})')
    full_path_items.append(f'/{node_label}')
    full_path = ''.join(full_path_items)
    return full_path


def descriptive_id_recursive_nodes(current_node: GraphNode, visited_nodes=None) -> List[GraphNode]:
    """ Returns descriptive id with nodes, not with its names. """
    if visited_nodes is None:
        visited_nodes = []

    full_path_items = []
    if current_node in visited_nodes:
        return []
    visited_nodes.append(current_node)
    if current_node.nodes_from:
        previous_items = []
        for parent_node in current_node.nodes_from:
            previous_items.extend(descriptive_id_recursive_nodes(parent_node, copy(visited_nodes)))
        full_path_items.extend(previous_items)
    full_path_items.append(current_node)
    return full_path_items



from abc import abstractmethod, ABC
from typing import Union, List, Optional, Sequence

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.objective import Objective
from golem.structural_analysis.graph_sa.entities.edge import Edge
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements


class BaseAnalyzeApproach(ABC):
    """
    Base class for analysis approach.
    :param graph: Graph containing the analyzed Node
    :param objective: objective functions for computing metric values
    """

    def __init__(self, graph: Graph, objective: Objective,
                 requirements: StructuralAnalysisRequirements = None):
        self._graph = graph
        self._objective = objective
        self._origin_metrics = list()
        self._requirements = \
            StructuralAnalysisRequirements() if requirements is None else requirements

    @abstractmethod
    def analyze(self, entity: Union[GraphNode, Edge], **kwargs) -> Union[List[dict], List[float]]:
        """ Creates the difference metric(scorer, index, etc) of the changed
        graph in relation to the original one.
        :param entity: entity to analyze.
        """
        pass

    @abstractmethod
    def sample(self, *args) -> Union[List[Graph], Graph]:
        """ Changes the graph according to the approach. """
        pass

    @staticmethod
    def _compare_with_origin_by_metric(origin_metric: Optional[float], modified_metric: Optional[float]) -> float:
        """ Calculates one metric. """
        if not modified_metric or not origin_metric:
            return -1.0

        if modified_metric < 0.0:
            numerator = modified_metric
            denominator = origin_metric
        else:
            numerator = origin_metric
            denominator = modified_metric

        if denominator == 0:
            return -1.0

        return numerator / denominator

    def _compare_with_origin_by_metrics(self, modified_graph: Graph) -> List[float]:
        """ Returns all relative metrics calculated. """
        modified_graph_metrics = self._objective(modified_graph).values

        if not self._origin_metrics:
            self._origin_metrics = self._objective(self._graph).values

        res = []
        for i in range(len(modified_graph_metrics)):
            res.append(self._compare_with_origin_by_metric(modified_metric=modified_graph_metrics[i],
                                                           origin_metric=self._origin_metrics[i]))
        return res



from typing import Union, Sequence, List, Optional, Tuple, Type

from golem.core.dag.graph import Graph, ReconnectType
from golem.core.dag.graph_node import GraphNode
from golem.core.dag.linked_graph import LinkedGraph


class GraphDelegate(Graph):
    """
    Graph that delegates calls to another Graph implementation.

    The class purpose is for cleaner code organisation:
    - avoid inheriting from specific Graph implementations
    - hide Graph implementation details from inheritors.
    """

    def __init__(self, *args, delegate_cls: Type[Graph] = LinkedGraph, **kwargs):
        self.operator = delegate_cls(*args, **kwargs)

    def add_node(self, node: GraphNode):
        self.operator.add_node(node)

    def update_node(self, old_node: GraphNode, new_node: GraphNode):
        self.operator.update_node(old_node, new_node)

    def update_subtree(self, old_subtree: GraphNode, new_subtree: GraphNode):
        self.operator.update_subtree(old_subtree, new_subtree)

    def delete_node(self, node: GraphNode, reconnect: ReconnectType = ReconnectType.single):
        self.operator.delete_node(node, reconnect)

    def delete_subtree(self, subtree: GraphNode):
        self.operator.delete_subtree(subtree)

    def node_children(self, node: GraphNode) -> Sequence[Optional[GraphNode]]:
        return self.operator.node_children(node=node)

    def connect_nodes(self, node_parent: GraphNode, node_child: GraphNode):
        self.operator.connect_nodes(node_parent, node_child)

    def disconnect_nodes(self, node_parent: GraphNode, node_child: GraphNode,
                         clean_up_leftovers: bool = False):
        self.operator.disconnect_nodes(node_parent, node_child, clean_up_leftovers)

    def get_edges(self) -> Sequence[Tuple[GraphNode, GraphNode]]:
        return self.operator.get_edges()

    def __eq__(self, other) -> bool:
        return self.operator.__eq__(other)

    def __str__(self):
        return self.operator.__str__()

    def __repr__(self):
        return self.operator.__repr__()

    def root_nodes(self) -> Sequence[GraphNode]:
        return self.operator.root_nodes()

    @property
    def root_node(self) -> Union[GraphNode, Sequence[GraphNode]]:
        return self.operator.root_node

    @property
    def nodes(self) -> List[GraphNode]:
        return self.operator.nodes

    @nodes.setter
    def nodes(self, new_nodes: List[GraphNode]):
        self.operator.nodes = new_nodes

    @property
    def descriptive_id(self):
        return self.operator.descriptive_id

    @property
    def length(self) -> int:
        return self.operator.length

    @property
    def depth(self) -> int:
        return self.operator.depth



import functools
import math
from copy import copy
from random import choice, randint, sample
from typing import Callable, List, Optional

from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.utilities.data_structures import ComparableEnum as Enum


class SelectionTypesEnum(Enum):
    tournament = 'tournament'
    spea2 = 'spea2'


class Selection(Operator):
    def __call__(self, population: PopulationT, pop_size: Optional[int] = None) -> PopulationT:
        """
        Selection of individuals based on specified type of selection
        :param population: A list of individuals to select from.
        :param pop_size: Optional custom population_size.
        Taken from algorithm parameters if not specified.
        """
        pop_size = pop_size or self.parameters.pop_size
        selection_type = choice(self.parameters.selection_types)
        return self._selection_by_type(selection_type)(population, pop_size)

    @staticmethod
    def _selection_by_type(selection_type: SelectionTypesEnum) -> Callable[[PopulationT, int], PopulationT]:
        selections = {
            SelectionTypesEnum.tournament: tournament_selection,
            SelectionTypesEnum.spea2: spea2_selection
        }
        if selection_type in selections:
            return selections[selection_type]
        else:
            raise ValueError(f'Required selection not found: {selection_type}')


def default_selection_behaviour(selection_func: Optional[Callable] = None, *, ensure_unique: bool = True,
                                populate_by_single: bool = True):
    def func_wrapper(func: Callable):
        @functools.wraps(func)
        def wrapper(individuals: PopulationT, pop_size: int, *args, **kwargs):
            if ensure_unique:
                individuals = list({ind.uid: ind for ind in individuals}.values())
            else:
                individuals = copy(individuals)

            if populate_by_single and len(individuals) == 1:
                return individuals * pop_size

            if len(individuals) <= pop_size:
                return individuals

            return func(individuals, pop_size, *args, **kwargs)
        return wrapper

    if selection_func:
        return func_wrapper(selection_func)  # Allows to decorate without args.
    return func_wrapper  # Allows to decorate with args but no selection_func specified.


@default_selection_behaviour
def tournament_selection(individuals: PopulationT, pop_size: int, fraction: float = 0.1) -> PopulationT:
    """ Having the size of *individuals* equals to *n* will have no effect other
    than lax ordering of *individuals*. """
    individuals = list(individuals)  # don't modify original
    group_size = math.ceil(len(individuals) * fraction)
    min_group_size = min(2, len(individuals))
    group_size = max(group_size, min_group_size)
    chosen = []
    iterations_limit = pop_size * 10
    for _ in range(iterations_limit):
        if len(chosen) >= pop_size:
            break
        group = sample(individuals, min(group_size, len(individuals)))
        best = max(group, key=lambda ind: ind.fitness)
        individuals.remove(best)
        chosen.append(best)
    return chosen


@default_selection_behaviour
def random_selection(individuals: PopulationT, pop_size: int) -> PopulationT:
    return sample(individuals, pop_size)


# Code of spea2 selection is modified part of DEAP library (Library URL: https://github.com/DEAP/deap).
@default_selection_behaviour
def spea2_selection(individuals: PopulationT, pop_size: int) -> PopulationT:
    """
    Apply SPEA-II selection operator on the *individuals*. Usually, the
    size of *individuals* will be larger than *n* because any individual
    present in *individuals* will appear in the returned list at most once.
    Having the size of *individuals* equals to *n* will have no effect other
    than sorting the population according to a strength Pareto scheme. The
    list returned contains references to the input *individuals*.

    :param individuals: A list of individuals to select from.
    :returns: A list of selected individuals
    """
    inds_len = len(individuals)
    fitness_len = len(individuals[0].fitness.values)
    inds_len_sqrt = math.sqrt(inds_len)
    strength_fits = [0] * inds_len
    fits = [0] * inds_len
    dominating_inds = [list() for _ in range(inds_len)]

    for i, ind_i in enumerate(individuals):
        for j, ind_j in enumerate(individuals[i + 1:], i + 1):
            if ind_i.fitness.dominates(ind_j.fitness):
                strength_fits[i] += 1
                dominating_inds[j].append(i)
            elif ind_j.fitness.dominates(ind_i.fitness):
                strength_fits[j] += 1
                dominating_inds[i].append(j)

    for i in range(inds_len):
        for j in dominating_inds[i]:
            fits[i] += strength_fits[j]

    # Choose all non-dominated individuals
    chosen_indices = [i for i in range(inds_len) if fits[i] < 1]

    if len(chosen_indices) < pop_size:  # The archive is too small
        for i in range(inds_len):
            distances = [0.0] * inds_len
            for j in range(i + 1, inds_len):
                dist = 0.0
                for idx in range(fitness_len):
                    val = \
                        individuals[i].fitness.values[idx] - \
                        individuals[j].fitness.values[idx]
                    dist += val * val
                distances[j] = dist
            kth_dist = _randomized_select(distances, 0, inds_len - 1, inds_len_sqrt)
            density = 1.0 / (kth_dist + 2.0)
            fits[i] += density

        next_indices = [(fits[i], i) for i in range(inds_len)
                        if i not in chosen_indices]
        next_indices.sort()
        # print next_indices
        chosen_indices += [i for _, i in next_indices[:pop_size - len(chosen_indices)]]

    elif len(chosen_indices) > pop_size:  # The archive is too large
        inds_len = len(chosen_indices)
        distances = [[0.0] * inds_len for _ in range(inds_len)]
        sorted_indices = [[0] * inds_len for _ in range(inds_len)]
        for i in range(inds_len):
            for j in range(i + 1, inds_len):
                dist = 0.0
                for idx in range(fitness_len):
                    val = \
                        individuals[chosen_indices[i]].fitness.values[idx] - \
                        individuals[chosen_indices[j]].fitness.values[idx]
                    dist += val * val
                distances[i][j] = dist
                distances[j][i] = dist
            distances[i][i] = -1

        # Insert sort is faster than quick sort for short arrays
        for i in range(inds_len):
            for j in range(1, inds_len):
                idx = j
                while idx > 0 and distances[i][j] < distances[i][sorted_indices[i][idx - 1]]:
                    sorted_indices[i][idx] = sorted_indices[i][idx - 1]
                    idx -= 1
                sorted_indices[i][idx] = j

        size = inds_len
        to_remove = []
        while size > pop_size:
            # Search for minimal distance
            min_pos = 0
            for i in range(1, inds_len):
                for j in range(1, size):
                    dist_i_sorted_j = distances[i][sorted_indices[i][j]]
                    dist_min_sorted_j = distances[min_pos][sorted_indices[min_pos][j]]

                    if dist_i_sorted_j < dist_min_sorted_j:
                        min_pos = i
                        break
                    elif dist_i_sorted_j > dist_min_sorted_j:
                        break

            # Remove minimal distance from sorted_indices
            for i in range(inds_len):
                distances[i][min_pos] = float("inf")
                distances[min_pos][i] = float("inf")

                for j in range(1, size - 1):
                    if sorted_indices[i][j] == min_pos:
                        sorted_indices[i][j] = sorted_indices[i][j + 1]
                        sorted_indices[i][j + 1] = min_pos

            # Remove corresponding individual from chosen_indices
            to_remove.append(min_pos)
            size -= 1

        for index in reversed(sorted(to_remove)):
            del chosen_indices[index]

    return [individuals[i] for i in chosen_indices]


# Auxiliary algorithmic functions for spea2_selection
# This code is a part of DEAP library (Library URL: https://github.com/DEAP/deap).
def _randomized_select(array: List[float], begin: int, end: int, i: float) -> float:
    """
    Allows to select the ith smallest element from array without sorting it.
    Runtime is expected to be O(n).
    """
    if begin == end:
        return array[begin]
    q = _randomized_partition(array, begin, end)
    k = q - begin + 1
    if i < k:
        return _randomized_select(array, begin, q, i)
    else:
        return _randomized_select(array, q + 1, end, i - k)


def _randomized_partition(array: List[float], begin: int, end: int) -> int:
    i = randint(begin, end)
    array[begin], array[i] = array[i], array[begin]
    return _partition(array, begin, end)


def _partition(array: List[float], begin: int, end: int) -> int:
    x = array[begin]
    i = begin - 1
    j = end + 1
    while True:
        j -= 1
        while array[j] > x:
            j -= 1
        i += 1
        while array[i] < x:
            i += 1
        if i < j:
            array[i], array[j] = array[j], array[i]
        else:
            return j



import os
from pathlib import Path
from typing import List, Optional, Sequence, Union

import numpy as np
import seaborn as sns
from matplotlib import cm, animation, pyplot as plt
from matplotlib.colors import Normalize

from golem.visualisation.opt_history.history_visualization import HistoryVisualization
from golem.visualisation.opt_history.utils import get_history_dataframe, get_description_of_operations_by_tag, \
    TagOperationsMap, LabelsColorMapType


class OperationsAnimatedBar(HistoryVisualization):
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None, dpi: Optional[int] = None,
                  best_fraction: Optional[float] = None, show_fitness: Optional[bool] = None,
                  tags_map: TagOperationsMap = None, palette: Optional[LabelsColorMapType] = None):
        """ Visualizes operations used across generations in the form of animated bar plot.

        :param save_path: path to save the visualization.
        :param dpi: DPI of the output figure.
        :param best_fraction: fraction of the best individuals of each generation that included in the
            visualization. Must be in the interval (0, 1].
        :param show_fitness: if False, the bar colors will not correspond to fitness.
        :param tags_map: if specified, all operations in the history are grouped based on the provided tags.
            If None, operations are not grouped.
        :param palette: a map from operation label to its color. If None, colors are picked by fixed colormap
            for every history independently.

        """

        save_path = save_path or self.get_predefined_value('save_path') or 'history_animated_bars.gif'
        dpi = dpi or self.get_predefined_value('dpi')
        best_fraction = best_fraction or self.get_predefined_value('best_fraction')
        show_fitness = show_fitness if show_fitness is not None else self.get_predefined_value('show_fitness') or True
        tags_map = tags_map or self.visuals_params.get('tags_map')
        palette = palette or self.visuals_params.get('palette')

        def interpolate_points(point_1, point_2, smoothness=18, power=4) -> List[np.array]:
            t_interp = np.linspace(0, 1, smoothness)
            point_1, point_2 = np.array(point_1), np.array(point_2)
            return [point_1 * (1 - t ** power) + point_2 * t ** power for t in t_interp]

        def smoothen_frames_data(data: Sequence[Sequence['ArrayLike']], smoothness=18, power=4) -> List[np.array]:
            final_frames = []
            for initial_frame in range(len(data) - 1):
                final_frames += interpolate_points(data[initial_frame], data[initial_frame + 1], smoothness, power)
            # final frame interpolates into itcls
            final_frames += interpolate_points(data[-1], data[-1], smoothness, power)

            return final_frames

        def animate(frame_num):
            frame_count = bar_data[frame_num]
            frame_color = bar_color[frame_num] if show_fitness else None
            frame_title = bar_title[frame_num]

            plt.title(frame_title)
            for bar_num in range(len(bars)):
                bars[bar_num].set_width(frame_count[bar_num])
                if not show_fitness:
                    continue
                bars[bar_num].set_facecolor(frame_color[bar_num])

        save_path = Path(save_path)
        if save_path.suffix not in ['.gif', '.mp4']:
            raise ValueError('A correct file extension (".mp4" or ".gif") should be set to save the animation.')

        animation_frames_per_step = 18
        animation_interval_between_frames_ms = 40
        animation_interpolation_power = 4
        fitness_colormap = cm.get_cmap('YlOrRd')

        generation_column_name = 'Generation'
        fitness_column_name = 'Fitness'
        operation_column_name = 'Operation'
        column_for_operation = 'tag' if tags_map else 'node'

        df_history = get_history_dataframe(self.history, best_fraction, tags_map)
        df_history = df_history.rename({
            'generation': generation_column_name,
            'fitness': fitness_column_name,
            column_for_operation: operation_column_name,
        }, axis='columns')
        operations_found = df_history[operation_column_name].unique()
        if tags_map:
            tags_all = list(tags_map.keys())
            operations_found = [tag for tag in tags_all if tag in operations_found]
            nodes_per_tag = df_history.groupby(operation_column_name)['node'].unique()
            bars_labels = [get_description_of_operations_by_tag(t, nodes_per_tag[t], 22) for t in operations_found]
        else:
            bars_labels = operations_found

        if palette:
            no_fitness_palette = palette
        else:
            no_fitness_palette = sns.color_palette('tab10', n_colors=len(operations_found))
            no_fitness_palette = {o: no_fitness_palette[i] for i, o in enumerate(operations_found)}

        # Getting normed fraction of individuals  per generation that contain operations given.
        generation_sizes = df_history.groupby(generation_column_name)['individual'].nunique()
        operations_with_individuals_count = df_history.groupby(
            [generation_column_name, operation_column_name],
            as_index=False
        ).aggregate({'individual': 'nunique'})
        operations_with_individuals_count['individual'] = operations_with_individuals_count.apply(
            lambda row: row['individual'] / generation_sizes[row[generation_column_name]],
            axis='columns')

        if show_fitness:
            # Getting fitness per individual with the list of contained operations in the form of
            # '.operation_1.operation_2. ... .operation_n.'
            individuals_fitness = df_history[[generation_column_name, 'individual', fitness_column_name]] \
                .drop_duplicates()
            individuals_fitness['operations'] = individuals_fitness.apply(
                lambda row: '.{}.'.format('.'.join(
                    df_history[
                        (df_history[generation_column_name] == row[generation_column_name]) &
                        (df_history['individual'] == row['individual'])
                        ][operation_column_name])),
                axis='columns')
            # Getting mean fitness of individuals with the operations given.
            operations_with_individuals_count[fitness_column_name] = operations_with_individuals_count.apply(
                lambda row: individuals_fitness[
                    (individuals_fitness[generation_column_name] == row[generation_column_name]) &
                    (individuals_fitness['operations'].str.contains(f'.{row[operation_column_name]}.'))
                    ][fitness_column_name].mean(),
                axis='columns')
            del individuals_fitness
        # Replacing the initial DataFrame with the processed one
        df_history = operations_with_individuals_count.set_index([generation_column_name, operation_column_name])
        del operations_with_individuals_count

        min_fitness = df_history[fitness_column_name].min() if show_fitness else None
        max_fitness = df_history[fitness_column_name].max() if show_fitness else None

        generations = generation_sizes.index.unique()
        bar_data = []
        bar_color = []
        # Getting data by tags through all generations and filling with zeroes where no such tag
        for gen_num in generations:
            bar_data.append([df_history.loc[gen_num]['individual'].get(tag, 0) for tag in operations_found])
            if not show_fitness:
                continue
            fitnesses = [df_history.loc[gen_num][fitness_column_name].get(tag, 0) for tag in operations_found]
            # Transfer fitness to color
            bar_color.append([
                fitness_colormap((fitness - min_fitness) / (max_fitness - min_fitness)) for fitness in fitnesses])

        bar_data = smoothen_frames_data(bar_data, animation_frames_per_step, animation_interpolation_power)
        title_template = 'Generation {}'
        if best_fraction is not None:
            title_template += f', top {best_fraction * 100}%'
        bar_title = [i for gen_num in generations for i in [title_template.format(gen_num)] * animation_frames_per_step]

        fig, ax = plt.subplots(figsize=(8, 5), facecolor='w')
        if show_fitness:
            bar_color = smoothen_frames_data(bar_color, animation_frames_per_step, animation_interpolation_power)
            sm = cm.ScalarMappable(norm=Normalize(min_fitness, max_fitness), cmap=fitness_colormap)
            sm.set_array([])
            fig.colorbar(sm, label=fitness_column_name, ax=ax)

        count = bar_data[0]
        color = bar_color[0] if show_fitness else [no_fitness_palette[tag] for tag in operations_found]
        title = bar_title[0]

        label_size = 10
        if any(len(label.split('\n')) > 2 for label in bars_labels):
            label_size = 8

        bars = ax.barh(bars_labels, count, color=color)
        ax.tick_params(axis='y', which='major', labelsize=label_size)
        ax.set_title(title)
        ax.set_xlim(0, 1)
        ax.set_xlabel(f'Fraction of graphs containing the operation')
        ax.xaxis.grid(True)
        ax.set_ylabel(operation_column_name)
        ax.invert_yaxis()
        plt.tight_layout()

        if not save_path.is_absolute():
            save_path = Path.cwd().joinpath(save_path)

        ani = animation.FuncAnimation(
            fig,
            animate,
            frames=len(bar_data),
            interval=animation_interval_between_frames_ms,
            repeat=True
        )
        ani.save(str(save_path), dpi=dpi)
        self.log.info(f'The animation was saved to "{save_path}".')
        plt.close(fig=fig)



from __future__ import annotations

import os
from pathlib import Path
from textwrap import wrap
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Union

import pandas as pd
from matplotlib import pyplot as plt

from golem.core.log import default_log

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.opt_history import OptHistory

MatplotlibColorType = Union[str, Sequence[float]]
LabelsColorMapType = Dict[str, MatplotlibColorType]

TagOperationsMap = Dict[str, List[str]]


def get_history_dataframe(history: OptHistory, best_fraction: Optional[float] = None,
                          tags_map: Optional[TagOperationsMap] = None):
    history_data = {
        'generation': [],
        'individual': [],
        'fitness': [],
        'node': [],
    }
    if tags_map:
        history_data['tag'] = []
        new_map = {}
        for tag, operations in tags_map.items():
            new_map.update({operation: tag for operation in operations})
        tags_map = new_map

    uid_counts = {}  # Resolving individuals with the same uid
    for gen_num, gen in enumerate(history.generations):
        for ind in gen:
            uid_counts[ind.uid] = uid_counts.get(ind.uid, -1) + 1
            for node in ind.graph.nodes:
                history_data['generation'].append(gen_num)
                history_data['individual'].append('_'.join([ind.uid, str(uid_counts[ind.uid])]))
                fitness = abs(ind.fitness.value)
                history_data['fitness'].append(fitness)
                history_data['node'].append(str(node))
                if not tags_map:
                    continue
                history_data['tag'].append(tags_map.get(str(node), None))

    df_history = pd.DataFrame.from_dict(history_data)

    if best_fraction is not None:
        generation_sizes = df_history.groupby('generation')['individual'].nunique()

        df_individuals = df_history[['generation', 'individual', 'fitness']] \
            .drop_duplicates(ignore_index=True)

        df_individuals['rank_per_generation'] = df_individuals.sort_values('fitness', ascending=False). \
            groupby('generation').cumcount()

        best_individuals = df_individuals[
            df_individuals.apply(
                lambda row: row['rank_per_generation'] < generation_sizes[row['generation']] * best_fraction,
                axis='columns'
            )
        ]['individual']

        df_history = df_history[df_history['individual'].isin(best_individuals)]

    return df_history


def get_description_of_operations_by_tag(tag: str, operations_by_tag: List[str], max_line_length: int,
                                         format_tag: str = 'it'):
    def make_text_fancy(text: str):
        return text.replace('_', ' ')

    def format_text(text_to_wrap: str, latex_format_tag: str = 'it') -> str:
        formatted_text = f'$\\{latex_format_tag}{{{text_to_wrap}}}$'
        formatted_text = formatted_text.replace(' ', '\\;')
        return formatted_text

    def format_wrapped_text_from_beginning(wrapped_text: List[str], part_to_format: str, latex_format_tag: str = 'it') \
            -> List[str]:
        for line_num, line in enumerate(wrapped_text):
            if part_to_format in line:
                # The line contains the whole part_to_format.
                wrapped_text[line_num] = line.replace(part_to_format, format_text(part_to_format, latex_format_tag))
                break
            if part_to_format.startswith(line):
                # The whole line should be formatted.
                wrapped_text[line_num] = format_text(line, latex_format_tag)
                part_to_format = part_to_format[len(line):].strip()

        return wrapped_text

    tag = make_text_fancy(tag)
    operations_by_tag = ', '.join(operations_by_tag)
    description = f'{tag}: {operations_by_tag}.'
    description = make_text_fancy(description)
    description = wrap(description, max_line_length)
    description = format_wrapped_text_from_beginning(description, tag, format_tag)
    description = '\n'.join(description)
    return description


def show_or_save_figure(figure: plt.Figure, save_path: Optional[Union[os.PathLike, str]], dpi: int = 100):
    if not save_path:
        plt.show()
    else:
        save_path = Path(save_path)
        if not save_path.is_absolute():
            save_path = Path.cwd().joinpath(save_path)
        figure.savefig(save_path, dpi=dpi)
        default_log().info(f'The figure was saved to "{save_path}".')
        plt.close()



import os

from examples.synthetic_graph_evolution.graph_search import graph_search_setup
from experiments.experiment_launcher import GraphStructureExperimentLauncher
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.paths import project_root

if __name__ == '__main__':
    path_to_save = os.path.join(project_root(), 'examples', 'experiment_launcher', 'experiment_results')

    launcher = GraphStructureExperimentLauncher(optimizer_cls=EvoGraphOptimizer,
                                                graph_names=['gnp', 'tree'],
                                                graph_sizes=[5, 10],
                                                num_trials=2,
                                                trial_iterations=2,
                                                is_save_visualizations=True,
                                                path_to_save=path_to_save)

    launcher.launch(optimizer_setup=graph_search_setup)



from __future__ import annotations

import inspect

from typing import TYPE_CHECKING, Any, Callable, Dict, List

if TYPE_CHECKING:
    from golem.visualisation.opt_history.history_visualization import HistoryVisualization

ArgConstraintChecker = Callable[..., Dict[str, Any]]


def per_time(visualization: HistoryVisualization, **kwargs) -> Dict[str, Any]:
    value = kwargs.get('per_time')
    if value and visualization.history.generations[0][0].metadata.get('evaluation_time_iso') is None:
        visualization.log.warning('Evaluation time not found in optimization history. '
                                  'Showing fitness line per generations...')
        kwargs['per_time'] = False
    return kwargs


def best_fraction(visualization: HistoryVisualization, **kwargs) -> Dict[str, Any]:
    value = kwargs.get('best_fraction')
    if value is not None and not 0 < value <= 1:
        raise ValueError('`best_fraction` argument should be in the interval (0, 1].')
    return kwargs


class ArgConstraintWrapper(type):
    """ Metaclass that wraps '.visualize' method into a series of argument constraint checkers.

    The behavior of the program due to invalid keyword argument values is entirely up to the checkers.
    Constraint checkers must return `kwargs` that may be modified during the check.

    The `DEFAULT_CONSTRAINTS` attribute is a mapping of argument names to their default constraint functions
    that will apply to any visualizations with this argument.

    Note that the checkers have all the `kwargs` of `visualize()` fully available, including the default
    ones, as well as the HistoryVisualization instance itself.
    """
    DEFAULT_CONSTRAINTS = {
        'best_fraction': best_fraction,
        'per_time': per_time
    }

    @staticmethod
    def wrap_constraints(constraint_checkers: List[ArgConstraintChecker]):
        def decorator(visualize_function):
            """Return a wrapped instance method"""

            def outer(visualization, *args, **kwargs):
                visualization_parameters = inspect.signature(visualize_function).parameters
                # Turn args into kwargs.
                parameter_names = list(visualization_parameters.keys())[1:]  # Skip `self`.
                args = {parameter_names[arg_num]: arg_value for arg_num, arg_value in enumerate(args)}
                kwargs.update(args)
                # Get default values and add them to the kwargs.
                default_kwargs = {p_name: p.default for p_name, p in visualization_parameters.items()
                                  if p.default is not p.empty}
                default_kwargs.update(kwargs)
                kwargs = default_kwargs
                # Filter wrong kwargs with warning.
                kwargs_to_ignore = []
                for argument in kwargs.keys():
                    if argument not in visualization_parameters:
                        visualization.log.warning(
                            f'Argument `{argument}` is not supported for "{visualization.__class__.__name__}". '
                            f'It is ignored.')
                        kwargs_to_ignore.append(argument)
                kwargs = {key: value for key, value in kwargs.items() if key not in kwargs_to_ignore}
                # Apply constraint_checkers iteratively.
                for checker in constraint_checkers:
                    kwargs = checker(visualization, **kwargs)
                # Make a visualization.
                visualization.log.info(
                    'Visualizing optimization history... It may take some time, depending on the history size.')
                return_value = visualize_function(visualization, **kwargs)
                return return_value

            return outer

        return decorator

    def __new__(mcs, name, bases, attrs):
        """If the class has a 'visualize' method, wrap it"""
        constraint_checkers = []
        if 'visualize' in attrs:
            parameters = inspect.signature(attrs['visualize']).parameters
            for kwarg, constraint_checker in mcs.DEFAULT_CONSTRAINTS.items():
                if kwarg in parameters:
                    constraint_checkers.append(constraint_checker)
            if 'constraint_checkers' in attrs:
                # Class-defined checkers
                for constraint_checker in attrs['constraint_checkers']:
                    constraint_checkers.append(constraint_checker)
            # Pass the checkers (or empty list) to the wrapper anyway.
            attrs['visualize'] = mcs.wrap_constraints(constraint_checkers)(attrs['visualize'])

        return super(ArgConstraintWrapper, mcs).__new__(mcs, name, bases, attrs)



from enum import Enum
from typing import Union, Tuple

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode


class EntityTypesEnum(Enum):
    node = 'node'
    edge = 'edge'


def get_entity_from_str(graph: Graph, entity_str: str) -> Union[GraphNode, Tuple[GraphNode, GraphNode]]:
    """ Gets entity from entity str using graph. """
    if len(entity_str.split('_')) == 2:
        parent_node_idx, child_node_idx = entity_str.split('_')
        parent_node = graph.nodes[int(parent_node_idx)]
        child_node = graph.nodes[int(child_node_idx)]
        return parent_node, child_node
    else:
        return graph.nodes[int(entity_str)]



import os
from copy import deepcopy
from pathlib import Path
from typing import List, Optional, Tuple, Dict
import multiprocessing

from golem.core.log import default_log
from golem.core.dag.graph import Graph, GraphNode
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_node_factory import OptNodeFactory
from golem.core.optimisers.timer import OptimisationTimer
from golem.structural_analysis.graph_sa.edge_sa_approaches import EdgeAnalyzeApproach, EdgeDeletionAnalyze, \
    EdgeReplaceOperationAnalyze
from golem.structural_analysis.graph_sa.edges_analysis import EdgesAnalysis
from golem.structural_analysis.graph_sa.entities.edge import Edge
from golem.structural_analysis.graph_sa.node_sa_approaches import NodeAnalyzeApproach, NodeDeletionAnalyze, \
    NodeReplaceOperationAnalyze, SubtreeDeletionAnalyze
from golem.structural_analysis.graph_sa.nodes_analysis import NodesAnalysis
from golem.structural_analysis.graph_sa.results.object_sa_result import ObjectSAResult
from golem.structural_analysis.graph_sa.results.sa_analysis_results import SAAnalysisResults
from golem.structural_analysis.graph_sa.results.utils import EntityTypesEnum
from golem.structural_analysis.graph_sa.sa_approaches_repository import StructuralAnalysisApproachesRepository
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements
from golem.visualisation.graph_viz import NodeColorType


class GraphStructuralAnalysis:
    """
    This class works as facade and allows to apply all kind of approaches
    to whole graph and separate nodes together.

    :param objective: list of objective functions for computing metric values
    :param node_factory: node factory to advise changes from available operations and models
    :param approaches: methods applied to graph. Default: None
    :param requirements: extra requirements to define specific details for different approaches.\
    See StructuralAnalysisRequirements class documentation.
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural/
    Default: False
    """

    def __init__(self, objective: Objective,
                 node_factory: OptNodeFactory,
                 is_preproc: bool = True,
                 approaches: List = None,
                 requirements: StructuralAnalysisRequirements = StructuralAnalysisRequirements(),
                 path_to_save: str = None,
                 is_visualize_per_iteration: bool = False):

        self.is_preproc = is_preproc
        self._log = default_log(self)

        if approaches:
            self.nodes_analyze_approaches = [approach for approach in approaches
                                             if issubclass(approach, NodeAnalyzeApproach)]
            self.edges_analyze_approaches = [approach for approach in approaches
                                             if issubclass(approach, EdgeAnalyzeApproach)]
        else:
            self._log.message('Approaches for analysis are not given, thus will be set to defaults.')
            self.nodes_analyze_approaches = [NodeDeletionAnalyze, NodeReplaceOperationAnalyze,
                                             SubtreeDeletionAnalyze]
            self.edges_analyze_approaches = [EdgeDeletionAnalyze, EdgeReplaceOperationAnalyze]

        self._nodes_analyze = NodesAnalysis(objective=objective,
                                            node_factory=node_factory,
                                            approaches=self.nodes_analyze_approaches,
                                            requirements=requirements,
                                            path_to_save=path_to_save)

        self._edges_analyze = EdgesAnalysis(objective=objective,
                                            approaches=self.edges_analyze_approaches,
                                            requirements=requirements,
                                            path_to_save=path_to_save)

        self.main_metric_idx = requirements.main_metric_idx
        self.path_to_save = path_to_save
        self.is_visualize_per_iteration = is_visualize_per_iteration

    def analyze(self, graph: Graph,
                result: SAAnalysisResults = None,
                nodes_to_analyze: List[GraphNode] = None, edges_to_analyze: List[Edge] = None,
                n_jobs: int = 1, timer: OptimisationTimer = None) -> SAAnalysisResults:
        """
        Applies defined structural analysis approaches

        :param graph: graph object to analyze
        :param result: analysis result
        :param nodes_to_analyze: nodes to analyze. Default: all nodes
        :param edges_to_analyze: edges to analyze. Default: all edges
        :param n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's).
        Tip: if specified graph isn't huge (as NN, for example) than set n_jobs to default value.
        :param timer: timer with timeout left for optimization
        """

        if not result:
            result = SAAnalysisResults()

        if n_jobs == -1:
            n_jobs = multiprocessing.cpu_count()

        if self.is_preproc:
            graph = self.graph_preprocessing(graph=graph)

        if self.nodes_analyze_approaches:
            self._nodes_analyze.analyze(graph=graph,
                                        results=result,
                                        nodes_to_analyze=nodes_to_analyze,
                                        n_jobs=n_jobs, timer=timer)

        if self.edges_analyze_approaches:
            self._edges_analyze.analyze(graph=graph,
                                        results=result,
                                        edges_to_analyze=edges_to_analyze,
                                        n_jobs=n_jobs, timer=timer)

        return result

    def optimize(self, graph: Graph,
                 n_jobs: int = 1, timer: OptimisationTimer = None,
                 max_iter: int = 10) -> Tuple[Graph, SAAnalysisResults]:
        """ Optimizes graph by applying 'analyze' method and deleting/replacing parts
        of graph iteratively
        :param graph: graph object to analyze.
        :param n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's).
        Tip: if specified graph isn't huge (as NN, for example) than set n_jobs to default value.
        :param timer: timer with timeout left for optimization.
        :param max_iter: max number of iterations of analysis. """

        approaches_repo = StructuralAnalysisApproachesRepository()
        approaches = self._nodes_analyze.approaches + self._edges_analyze.approaches
        approaches_names = [approach.__name__ for approach in approaches]

        # what actions were applied on the graph and how many
        actions_applied = dict.fromkeys(approaches_names, 0)

        result = SAAnalysisResults()

        analysis_result = self.analyze(graph=graph, result=result, timer=timer, n_jobs=n_jobs)
        converged = False
        iter = 0

        if analysis_result.is_empty:
            self._log.message(f'{iter} actions were taken during SA')
            return graph, analysis_result

        while not converged:
            iter += 1
            worst_result = analysis_result.get_info_about_worst_result(
                metric_idx_to_optimize_by=self.main_metric_idx)
            if self.is_visualize_per_iteration:
                self.visualize_on_graph(graph=deepcopy(graph), analysis_result=analysis_result,
                                        metric_idx_to_optimize_by=self.main_metric_idx,
                                        mode='final',
                                        font_size_scale=0.6)
            if worst_result['value'] > 1:
                # apply the worst approach
                postproc_method = approaches_repo.postproc_method_by_name(worst_result['approach_name'])
                graph = postproc_method(graph=graph, worst_result=worst_result)
                actions_applied[f'{worst_result["approach_name"]}'] += 1

                if timer is not None and timer.is_time_limit_reached():
                    break

                if max_iter and iter >= max_iter:
                    break

                analysis_result = self.analyze(graph=graph,
                                               result=result,
                                               n_jobs=n_jobs,
                                               timer=timer)
            else:
                converged = True

        self._log.message(f'{iter} iterations passed during SA')
        self._log.message(f'The following actions were applied during SA: {actions_applied}')

        if self.path_to_save:
            if not os.path.exists(self.path_to_save):
                os.makedirs(self.path_to_save)
            analysis_result.save(path=self.path_to_save)

        return graph, analysis_result

    @staticmethod
    def apply_results(graph: Graph, analysis_result: SAAnalysisResults,
                      metric_idx_to_optimize_by: int, iter: int = None) -> Graph:
        """ Optimizes graph by applying actions specified in analysis_result. """

        def optimize_on_iter(graph: Graph, analysis_result: SAAnalysisResults,
                             metric_idx_to_optimize_by: int, iter: int = None) -> Graph:
            """ Get worst result on specified iteration and process graph with it. """
            worst_result = analysis_result.get_info_about_worst_result(
                metric_idx_to_optimize_by=metric_idx_to_optimize_by, iter=iter)
            approaches_repo = StructuralAnalysisApproachesRepository()
            postproc_method = approaches_repo.postproc_method_by_name(worst_result['approach_name'])
            graph = postproc_method(graph=graph, worst_result=worst_result)
            return graph

        if iter is not None:
            return optimize_on_iter(graph=graph, analysis_result=analysis_result,
                                    metric_idx_to_optimize_by=metric_idx_to_optimize_by, iter=iter)

        num_of_iter = len(analysis_result.results_per_iteration)
        for i in range(num_of_iter):
            graph = optimize_on_iter(graph=graph, analysis_result=analysis_result,
                                     metric_idx_to_optimize_by=metric_idx_to_optimize_by, iter=iter)
        return graph

    @staticmethod
    def visualize_on_graph(graph: Graph, analysis_result: SAAnalysisResults,
                           metric_idx_to_optimize_by: int, mode: str = 'final',
                           save_path: str = None, node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
                           node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
                           edge_curvature_scale: Optional[float] = None):
        """ Visualizes results of Structural Analysis on graph(s).
        :param graph: initial graph before SA
        :param analysis_result: results of Structural Analysis
        :param metric_idx_to_optimize_by: index of optimized metric
        :param mode: 'first' -- visualize only first iteration of SA,
                     'final' - visualize only the last iteration of SA,
                     'by_iteration' -- visualize every iteration of SA.
        :param save_path: path to save visualizations
        :param node_color: color of nodes to use.
        :param node_size_scale: use to make node size bigger or lesser. Supported only for the engine 'matplotlib'.
        :param font_size_scale: use to make font size bigger or lesser. Supported only for the engine 'matplotlib'.
        :param edge_curvature_scale: use to make edges more or less curved. Supported only for the engine 'matplotlib'.
        :param dpi: DPI of the output image. Not supported for the engine 'pyvis'.
        """

        def get_nodes_and_edges_labels(analysis_result: SAAnalysisResults, iter: int) \
                -> Tuple[Dict[int, str], Dict[int, str]]:
            """ Get nodes and edges labels in dictionary form. """

            def get_str_labels(result: ObjectSAResult) -> str:
                """ Get string results. """
                approaches = result.result_approaches
                cur_label = ''
                for approach in approaches:
                    approach_name = result._get_approach_name(approach=approach)
                    if 'del' in approach_name.lower():
                        short_approach_name = 'D'
                    else:
                        short_approach_name = 'R'
                    cur_label += f'{short_approach_name}: {approach.get_rounded_metrics(idx=2)}\n'
                return cur_label

            nodes_labels = {}
            for i, node_result in enumerate(analysis_result.results_per_iteration[iter][EntityTypesEnum.node.value]):
                nodes_labels[i] = get_str_labels(result=node_result)

            edges_labels = {}
            for i, edge_result in enumerate(analysis_result.results_per_iteration[iter][EntityTypesEnum.edge.value]):
                edges_labels[i] = get_str_labels(result=edge_result)

            return nodes_labels, edges_labels

        num_of_iter = len(analysis_result.results_per_iteration)

        if mode == 'first':
            iters = [0]
        else:
            iters = range(num_of_iter)

        for i in iters:
            nodes_labels, edges_labels = get_nodes_and_edges_labels(analysis_result=analysis_result, iter=i)
            if not (mode == 'final' and i != iters[-1]):
                if not Path.is_file(Path(save_path)):
                    j = 0
                    while f'pipeline_after_sa_{j}.png' in os.listdir(save_path):
                        j += 1
                    cur_path = os.path.join(save_path, f'pipeline_after_sa_{j}.png')
                graph.show(node_color=node_color, dpi=dpi, node_size_scale=node_size_scale,
                           nodes_labels=nodes_labels, font_size_scale=font_size_scale,
                           edge_curvature_scale=edge_curvature_scale,
                           edges_labels=edges_labels, save_path=cur_path)
                default_log("SA_visualization").info(f"SA visualization was saved to: {cur_path}")
            if mode == 'by_iteration':
                graph = GraphStructuralAnalysis.apply_results(graph=graph, analysis_result=analysis_result,
                                                              metric_idx_to_optimize_by=metric_idx_to_optimize_by,
                                                              iter=i)

    @staticmethod
    def graph_preprocessing(graph: Graph) -> Graph:
        """ Graph preprocessing, which consists in removing consecutive nodes
        with the same models/operations in the graph """
        for node_child in reversed(graph.nodes):
            if not node_child.nodes_from or len(node_child.nodes_from) != 1:
                continue
            nodes_uid_to_delete = []
            for node_parent in node_child.nodes_from:
                if node_child.name == node_parent.name:
                    nodes_uid_to_delete.append(node_parent.uid)
            # there is a need to store nodes using uid since after deleting one of the nodes in graph
            # other nodes will not remain the same (nodes_from may be changed)
            for uid in nodes_uid_to_delete:
                node_to_delete = [node for node in graph.nodes if node.uid == uid][0]
                graph.delete_node(node_to_delete)
        return graph



import os
from typing import Any, List, Optional

import numpy as np
import torch
from gensim.models import word2vec, Word2Vec
from mol2vec.features import mol2alt_sentence, MolSentence
from rdkit.Chem import AllChem, RDKFingerprint, rdFingerprintGenerator
from rdkit.ML.Descriptors.MoleculeDescriptors import MolecularDescriptorCalculator

from examples.molecule_search.mol_adapter import MolAdapter
from examples.molecule_search.mol_transformer.transformer import create_masks, Transformer, EXTRA_CHARS, ALPHABET_SIZE
from examples.molecule_search.utils import download_from_github
from golem.core.log import default_log
from golem.core.paths import project_root


def adapter_func_to_molgraph(func):
    """ Decorator function to adapt observation to MolGraphs graphs. """
    def wrapper(obs):
        mol_graph = MolAdapter().restore(obs)
        embedding = func(mol_graph)
        return embedding
    return wrapper


def adapter_method_to_molgraph(func):
    """ Decorator function to adapt observation to MolGraphs graphs. """
    def wrapper(obj, obs):
        mol_graph = MolAdapter().restore(obs)
        embedding = func(obj, mol_graph)
        return embedding
    return wrapper


@adapter_func_to_molgraph
def ECFP(obs: Any):
    """ Extended-Connectivity Fingerprint """
    molecule = obs.get_rw_molecule()
    feature_list = AllChem.GetMorganFingerprintAsBitVect(molecule,
                                                         radius=2,
                                                         nBits=2**10,
                                                         useFeatures=False,
                                                         useChirality=False)
    return np.array(feature_list)


@adapter_func_to_molgraph
def RDKF(obs: Any):
    """ RDK Fingerprint """
    molecule = obs.get_rw_molecule()
    fingerprint_rdk = RDKFingerprint(molecule)
    return np.array(fingerprint_rdk)


@adapter_func_to_molgraph
def atom_pair(obs: Any):
    """ Atom pair fingerprint """
    molecule = obs.get_rw_molecule()
    fingerprint = rdFingerprintGenerator.GetAtomPairGenerator(fpSize=1024).GetFingerprint(molecule)
    return np.array(fingerprint)


@adapter_func_to_molgraph
def topological_torsion(obs: Any):
    """ Topological Torsion fingerprint """
    molecule = obs.get_rw_molecule()
    fingerprint = rdFingerprintGenerator.GetTopologicalTorsionGenerator(fpSize=1024).GetFingerprint(molecule)
    return np.array(fingerprint)


@adapter_func_to_molgraph
def mol_descriptors(obs: Any):
    molecule = obs.get_rw_molecule()
    chosen_descriptors = ['BalabanJ', 'BertzCT', 'Chi0', 'Chi0n', 'Chi0v', 'Chi1', 'Chi1n', 'Chi1v', 'Chi2n', 'Chi2v',
                          'Chi3n', 'Chi3v', 'Chi4n', 'Chi4v', 'EState_VSA1', 'EState_VSA10', 'EState_VSA11',
                          'EState_VSA2', 'EState_VSA3', 'EState_VSA4', 'EState_VSA5', 'EState_VSA6', 'EState_VSA7',
                          'EState_VSA8', 'EState_VSA9', 'ExactMolWt', 'FpDensityMorgan1', 'FpDensityMorgan2',
                          'FpDensityMorgan3', 'FractionCSP3', 'HallKierAlpha', 'HeavyAtomCount', 'HeavyAtomMolWt',
                          'Ipc', 'Kappa1', 'Kappa2', 'Kappa3', 'LabuteASA', 'MaxAbsEStateIndex', 'MaxAbsPartialCharge',
                          'MaxEStateIndex', 'MaxPartialCharge', 'MinAbsEStateIndex', 'MinAbsPartialCharge',
                          'MinEStateIndex', 'MinPartialCharge', 'MolLogP', 'MolMR', 'MolWt', 'NHOHCount', 'NOCount',
                          'NumAliphaticCarbocycles', 'NumAliphaticHeterocycles', 'NumAliphaticRings',
                          'NumAromaticCarbocycles', 'NumAromaticHeterocycles', 'NumAromaticRings', 'NumHAcceptors',
                          'NumHDonors', 'NumHeteroatoms', 'NumRadicalElectrons', 'NumRotatableBonds',
                          'NumSaturatedCarbocycles', 'NumSaturatedHeterocycles', 'NumSaturatedRings',
                          'NumValenceElectrons', 'PEOE_VSA1', 'PEOE_VSA10', 'PEOE_VSA11', 'PEOE_VSA12',
                          'PEOE_VSA13', 'PEOE_VSA14', 'PEOE_VSA2', 'PEOE_VSA3', 'PEOE_VSA4', 'PEOE_VSA5',
                          'PEOE_VSA6', 'PEOE_VSA7', 'PEOE_VSA8', 'PEOE_VSA9', 'RingCount', 'SMR_VSA1', 'SMR_VSA10',
                          'SMR_VSA2', 'SMR_VSA3', 'SMR_VSA4', 'SMR_VSA5', 'SMR_VSA6', 'SMR_VSA7', 'SMR_VSA8',
                          'SMR_VSA9', 'SlogP_VSA1', 'SlogP_VSA10', 'SlogP_VSA11', 'SlogP_VSA12', 'SlogP_VSA2',
                          'SlogP_VSA3', 'SlogP_VSA4', 'SlogP_VSA5', 'SlogP_VSA6', 'SlogP_VSA7', 'SlogP_VSA8',
                          'SlogP_VSA9', 'TPSA', 'VSA_EState1', 'VSA_EState10', 'VSA_EState2', 'VSA_EState3',
                          'VSA_EState4', 'VSA_EState5', 'VSA_EState6', 'VSA_EState7', 'VSA_EState8', 'VSA_EState9',
                          'fr_Al_COO', 'fr_Al_OH', 'fr_Al_OH_noTert', 'fr_ArN', 'fr_Ar_COO', 'fr_Ar_N', 'fr_Ar_NH',
                          'fr_Ar_OH', 'fr_COO', 'fr_COO2', 'fr_C_O', 'fr_C_O_noCOO', 'fr_C_S', 'fr_HOCCN', 'fr_Imine',
                          'fr_NH0', 'fr_NH1', 'fr_NH2', 'fr_N_O', 'fr_Ndealkylation1', 'fr_Ndealkylation2',
                          'fr_Nhpyrrole', 'fr_SH', 'fr_aldehyde', 'fr_alkyl_carbamate', 'fr_alkyl_halide',
                          'fr_allylic_oxid', 'fr_amide', 'fr_amidine', 'fr_aniline', 'fr_aryl_methyl', 'fr_azide',
                          'fr_azo', 'fr_barbitur', 'fr_benzene', 'fr_benzodiazepine', 'fr_bicyclic', 'fr_diazo',
                          'fr_dihydropyridine', 'fr_epoxide', 'fr_ester', 'fr_ether', 'fr_furan', 'fr_guanido',
                          'fr_halogen', 'fr_hdrzine', 'fr_hdrzone', 'fr_imidazole', 'fr_imide', 'fr_isocyan',
                          'fr_isothiocyan', 'fr_ketone', 'fr_ketone_Topliss', 'fr_lactam', 'fr_lactone', 'fr_methoxy',
                          'fr_morpholine', 'fr_nitrile', 'fr_nitro', 'fr_nitro_arom', 'fr_nitro_arom_nonortho',
                          'fr_nitroso', 'fr_oxazole', 'fr_oxime', 'fr_para_hydroxylation', 'fr_phenol',
                          'fr_phenol_noOrthoHbond', 'fr_phos_acid', 'fr_phos_ester', 'fr_piperdine', 'fr_piperzine',
                          'fr_priamide', 'fr_prisulfonamd', 'fr_pyridine', 'fr_quatN', 'fr_sulfide', 'fr_sulfonamd',
                          'fr_sulfone', 'fr_term_acetylene', 'fr_tetrazole', 'fr_thiazole', 'fr_thiocyan',
                          'fr_thiophene', 'fr_unbrch_alkane', 'fr_urea', 'qed']
    mol_descriptor_calculator = MolecularDescriptorCalculator(chosen_descriptors)
    list_of_descriptor_vals = list(mol_descriptor_calculator.CalcDescriptors(molecule))
    return list_of_descriptor_vals


class Mol2Vec:

    PRETRAINED_WORD2VEC = 'examples/molecule_search/data/pretrained_models/model_300dim.pkl'
    GITHUB_URL = 'https://github.com/samoturk/mol2vec/raw/master/examples/models/model_300dim.pkl'

    def __init__(self):
        self.file_path = os.path.join(project_root(), Mol2Vec.PRETRAINED_WORD2VEC)
        download_from_github(self.file_path,
                             Mol2Vec.GITHUB_URL,
                             message="Downloading pretrained model for molecules encoding...")

        self.model = word2vec.Word2Vec.load(self.file_path)

    @adapter_method_to_molgraph
    def __call__(self, obs: Any):
        molecule = obs.get_rw_molecule()
        sentence = MolSentence(mol2alt_sentence(molecule, radius=1))
        embedding = self.sentences2vec([sentence], self.model, unseen='UNK')[0]
        return np.array(embedding).astype(float)

    @staticmethod
    def sentences2vec(sentences: List[MolSentence], model: Word2Vec, unseen: Optional[str] = None) -> np.array:
        """Generate vectors for each sentence (list) in a list of sentences. Vector is simply a
        sum of vectors for individual words.

        Parameters
        ----------
        sentences : list, array
            List with sentences
        model : word2vec.Word2Vec
            Gensim word2vec model
        unseen : None, str
            Keyword for unseen words. If None, those words are skipped.
            https://stats.stackexchange.com/questions/163005/how-to-set-the-dictionary-for-text-analysis-using-neural-networks/163032#163032

        Returns
        -------
        np.array
        """

        keys = set(model.wv.key_to_index)
        vec = []

        if unseen:
            unseen_vec = model.wv.get_vector(unseen)

        for sentence in sentences:
            if unseen:
                vec.append(sum([model.wv.get_vector(y) if y in set(sentence) & keys
                                else unseen_vec for y in sentence]))
            else:
                vec.append(sum([model.wv.get_vector(y) for y in sentence
                                if y in set(sentence) & keys]))
        return np.array(vec)


class MoleculeTransformer:
    """ Based on https://github.com/mpcrlab/MolecularTransformerEmbeddings """

    PRETRAINED_TRANSFORMER = 'examples/molecule_search/data/pretrained_models/pretrained.ckpt'
    GITHUB_URL = 'https://github.com/mpcrlab/MolecularTransformerEmbeddings/releases/download/' \
                 'checkpoints/pretrained.ckpt'

    def __init__(self, embedding_size: int = 512, num_layers: int = 6, max_length: int = 256):
        self.log = default_log(self)

        self.file_path = os.path.join(project_root(), MoleculeTransformer.PRETRAINED_TRANSFORMER)
        download_from_github(self.file_path,
                             MoleculeTransformer.GITHUB_URL,
                             message="Downloading pretrained model for molecules encoding...")
        self.model = self._model_setup(embedding_size, num_layers)
        self.encoder = self.model.encoder.cpu()
        self.max_length = max_length

    def _model_setup(self, embedding_size: int, num_layers: int):
        model = Transformer(ALPHABET_SIZE, embedding_size, num_layers).eval()
        model = torch.nn.DataParallel(model)
        checkpoint = torch.load(self.file_path, map_location=torch.device("cpu"))
        model.load_state_dict(checkpoint['state_dict'])
        return model.module.cpu()

    @adapter_method_to_molgraph
    def __call__(self, obs: Any):
        smiles = obs.get_smiles()
        with torch.no_grad():
            encoded = self.encode_smiles(smiles)
            mask = create_masks(encoded)
            embedding = self.encoder(encoded, mask)[0].numpy()
            embedding = embedding.mean(axis=0)
        return embedding

    @staticmethod
    def encode_char(c):
        return ord(c) - 32

    def encode_smiles(self, string: str, start_char=EXTRA_CHARS['seq_start']):
        return torch.tensor([ord(start_char)] +
                            [self.encode_char(c) for c in string], dtype=torch.long)[:self.max_length].unsqueeze(0)
